import re
from . import Parser, BlockState

# Import tokens in module namespace
from .tokens import (
    CommentToken,
    StringToken,
    UnterminatedStringToken,
    IdentifierToken,
    NonIdentifierToken,
    KeywordToken,
    BuiltinsToken,
    InstanceToken,
    NumberToken,
    FunctionNameToken,
    ClassNameToken,
    TodoCommentToken,
    OpenParenToken,
    CloseParenToken,
    IllegalToken,
)

# Keywords sets

# Source: import keyword; keyword.kwlist (Python 2.7)
python2Keywords = set(
    [
        "and",
        "as",
        "assert",
        "break",
        "class",
        "continue",
        "def",
        "del",
        "elif",
        "else",
        "except",
        "exec",
        "finally",
        "for",
        "from",
        "global",
        "if",
        "import",
        "in",
        "is",
        "lambda",
        "not",
        "or",
        "pass",
        "print",
        "raise",
        "return",
        "try",
        "while",
        "with",
        "yield",
    ]
)

# Source: import keyword; keyword.kwlist (Python 3.12)
python3Keywords = set(
    [
        "False",
        "None",
        "True",
        "and",
        "as",
        "assert",
        "break",
        "class",
        "continue",
        "def",
        "del",
        "elif",
        "else",
        "except",
        "finally",
        "for",
        "from",
        "global",
        "if",
        "import",
        "in",
        "is",
        "lambda",
        "nonlocal",
        "not",
        "or",
        "pass",
        "raise",
        "return",
        "try",
        "while",
        "with",
        "yield",
        "async",
        "await",
    ]
)

# Merge the two sets to get a general Python keyword list
pythonKeywords = python2Keywords | python3Keywords

# Builtins sets

# Source: dir (__builtins__) (Python 2.7.12)
python2Builtins = set(
    [
        "ArithmeticError",
        "AssertionError",
        "AttributeError",
        "BaseException",
        "BufferError",
        "BytesWarning",
        "DeprecationWarning",
        "EOFError",
        "Ellipsis",
        "EnvironmentError",
        "Exception",
        "False",
        "FloatingPointError",
        "FutureWarning",
        "GeneratorExit",
        "IOError",
        "ImportError",
        "ImportWarning",
        "IndentationError",
        "IndexError",
        "KeyError",
        "KeyboardInterrupt",
        "LookupError",
        "MemoryError",
        "NameError",
        "None",
        "NotImplemented",
        "NotImplementedError",
        "OSError",
        "OverflowError",
        "PendingDeprecationWarning",
        "ReferenceError",
        "RuntimeError",
        "RuntimeWarning",
        "StandardError",
        "StopIteration",
        "SyntaxError",
        "SyntaxWarning",
        "SystemError",
        "SystemExit",
        "TabError",
        "True",
        "TypeError",
        "UnboundLocalError",
        "UnicodeDecodeError",
        "UnicodeEncodeError",
        "UnicodeError",
        "UnicodeTranslateError",
        "UnicodeWarning",
        "UserWarning",
        "ValueError",
        "Warning",
        "ZeroDivisionError",
        "__debug__",
        "__doc__",
        "__import__",
        "__name__",
        "__package__",
        "abs",
        "all",
        "any",
        "apply",
        "basestring",
        "bin",
        "bool",
        "buffer",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "cmp",
        "coerce",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "execfile",
        "exit",
        "file",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "intern",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "long",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "quit",
        "range",
        "raw_input",
        "reduce",
        "reload",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "unichr",
        "unicode",
        "vars",
        "xrange",
        "zip",
    ]
)

# Source: import builtins; dir(builtins) (Python 3.12)
# Note: Removed 'False', 'None', 'True'. They are keyword in Python 3
python3Builtins = set(
    [
        "ArithmeticError",
        "AssertionError",
        "AttributeError",
        "BaseException",
        "BaseExceptionGroup",
        "BlockingIOError",
        "BrokenPipeError",
        "BufferError",
        "BytesWarning",
        "ChildProcessError",
        "ConnectionAbortedError",
        "ConnectionError",
        "ConnectionRefusedError",
        "ConnectionResetError",
        "DeprecationWarning",
        "EOFError",
        "Ellipsis",
        "EncodingWarning",
        "EnvironmentError",
        "Exception",
        "ExceptionGroup",
        "FileExistsError",
        "FileNotFoundError",
        "FloatingPointError",
        "FutureWarning",
        "GeneratorExit",
        "IOError",
        "ImportError",
        "ImportWarning",
        "IndentationError",
        "IndexError",
        "InterruptedError",
        "IsADirectoryError",
        "KeyError",
        "KeyboardInterrupt",
        "LookupError",
        "MemoryError",
        "ModuleNotFoundError",
        "NameError",
        "NotADirectoryError",
        "NotImplemented",
        "NotImplementedError",
        "OSError",
        "OverflowError",
        "PendingDeprecationWarning",
        "PermissionError",
        "ProcessLookupError",
        "RecursionError",
        "ReferenceError",
        "ResourceWarning",
        "RuntimeError",
        "RuntimeWarning",
        "StopAsyncIteration",
        "StopIteration",
        "SyntaxError",
        "SyntaxWarning",
        "SystemError",
        "SystemExit",
        "TabError",
        "TimeoutError",
        "TypeError",
        "UnboundLocalError",
        "UnicodeDecodeError",
        "UnicodeEncodeError",
        "UnicodeError",
        "UnicodeTranslateError",
        "UnicodeWarning",
        "UserWarning",
        "ValueError",
        "Warning",
        "ZeroDivisionError",
        "_",
        "__build_class__",
        "__debug__",
        "__doc__",
        "__import__",
        "__loader__",
        "__name__",
        "__package__",
        "__spec__",
        "abs",
        "aiter",
        "all",
        "anext",
        "any",
        "ascii",
        "bin",
        "bool",
        "breakpoint",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "exec",
        "exit",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "quit",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "vars",
        "zip",
    ]
)


# Merge the two sets to get a general Python builtins list
pythonBuiltins = python2Builtins | python3Builtins

# Instance sets
python2Instance = set(["self"])

python3Instance = set(["self"])

pythonInstance = python2Instance | python3Instance


class MultilineStringToken(StringToken):
    """Characters representing a multi-line string."""

    defaultStyle = "fore:#7F0000"


class CellCommentToken(CommentToken):
    """Characters representing a cell separator comment: "##"."""

    defaultStyle = "bold:yes, underline:yes"


stringLiteralPrefixes = frozenset("u|r|b|f|rb|br|rf|fr|t|rt|tr".split("|"))

# This regexp is used to find special stuff, such as comments, numbers and
# strings.
tokenProg = re.compile(
    "#|"  # Comment or
    + "("  # Begin of string group (group 1)
    + "("
    + "|".join(stringLiteralPrefixes)
    + ")?"  # (group 2)
    + "(\"\"\"|'''|\"|')"  # String start (triple quotes first, group 3)
    + ")|"  # End of string group
    + "([a-z0-9_]+)|"  # Identifiers/numbers (group 4) or
    + r"(\(|\[|\{)|"  # Opening parenthesis (gr 5)
    + r"(\)|\]|\})|"  # Closing parenthesis (gr 6)
    + "("
    + chr(160)  # non-breaking space (gr 7)
    + ")",
    re.IGNORECASE,
)


# For a given type of string ( ', " , ''' , """ ), get the RegExp
# program that matches the end. (^|[^\\]) means: start of the line
# or something that is not \ (since \ is supposed to escape the following
# quote) (\\\\)* means: any number of two slashes \\ since each slash will
# escape the next one
endProgs = {
    "'": re.compile(r"(^|[^\\])(\\\\)*'"),
    '"': re.compile(r'(^|[^\\])(\\\\)*"'),
    "'''": re.compile(r"(^|[^\\])(\\\\)*'''"),
    '"""': re.compile(r'(^|[^\\])(\\\\)*"""'),
}

# A string can also be line-continued with a backslash at the very end of a line.
# In that case one single or double quote sign can span more than a line.
stringLineContinuation = re.compile(r"(^|[^\\])(\\\\)*\\$")


class PythonParser(Parser):
    """Parser for Python in general."""

    _extensions = []
    _shebangKeywords = []
    # The list of keywords is overridden by the Python2/3 specific parsers
    _keywords = set()
    # The list of builtins and instances is overridden by the Python2/3 specific parsers
    _builtins = set()
    _instance = set()

    def _identifierState(self, identifier=None):
        """Given an identifier returns the identifier state:
        3 means the current identifier can be a function.
        4 means the current identifier can be a class.
        0 otherwise.

        This method enables storing the state during the line,
        and helps the Cython parser to reuse the Python parser's code.
        """
        if identifier is None:
            # Explicit get/reset
            try:
                state = self._idsState
            except Exception:
                state = 0
            self._idsState = 0
            return state
        elif identifier == "def":
            # Set function state
            self._idsState = 3
            return 3
        elif identifier == "class":
            # Set class state
            self._idsState = 4
            return 4
        else:
            # This one can be func or class, next one can't
            state = self._idsState
            self._idsState = 0
            return state

    def parseLine(self, line, previousState=0):
        """Parse a line of Python code, returning a list of tokens.
        previousState is the state of the previous block, and is used
        to handle line continuation and multiline strings.
        """

        # Init
        pos = 0  # Position following the previous match

        tokensForLine = []

        # identifierState and previousState values:
        # 0: nothing special
        # 1: multiline comment single qoutes
        # 2: multiline comment double quotes
        # 3: a def keyword
        # 4: a class keyword
        # 5: a single quote string literal (non-multiline) line-continued by a backslash
        # 6: a double quote string literal (non-multiline) line-continued by a backslash

        # Handle line continuation after def or class
        # identifierState is 3 or 4 if the previous identifier was 3 or 4
        if previousState == 3 or previousState == 4:
            self._identifierState({3: "def", 4: "class"}[previousState])
        else:
            self._identifierState(None)

        if previousState in [1, 2, 5, 6]:
            if previousState <= 2:
                token = MultilineStringToken(line, 0, 0)
            else:
                token = StringToken(line, 0, 0)
            token._style = ["", "'''", '"""', None, None, "'", '"'][previousState]
            tokens = self._findEndOfString(line, token)
            # Process tokens
            for token in tokens:
                tokensForLine.append(token)
                if isinstance(token, BlockState):
                    return tokensForLine
            pos = token.end

        # Enter the main loop that iterates over the tokens and skips strings
        while True:
            # Get next tokens
            tokens = self._findNextToken(line, pos)
            if not tokens:
                self._promoteMatchCaseSoftKeywords(tokensForLine)
                return tokensForLine
            elif isinstance(tokens[-1], StringToken):
                moreTokens = self._findEndOfString(line, tokens[-1])
                tokens = tokens[:-1] + moreTokens

            # Process tokens
            for token in tokens:
                tokensForLine.append(token)
                if isinstance(token, BlockState):
                    return tokensForLine
            pos = token.end

    @staticmethod
    def _promoteMatchCaseSoftKeywords(tokens):
        """promotes identifier tokens "match" and "case" to keyword tokens if appropriate

        list "tokens" contains the tokens of the current line

        A simple algorithm will be used that only knows about the tokens of the current
        line, but not about the lines before or after.
        If "match" or "case" is a keyword, its token in list "tokens" will be replaced by
        a keyword token.

        "match" or "case" will be considered a keyword if one of the two patterns is found:

        1.
            optional whitespace
            identifier token "match" or "case"
            all parenthesis like tokens ([{}]) must be properly closed
            colon (with optional whitespace)
            optional comment

        2.
            optional whitespace
            identifier token "match" or "case"
            parenthesis like tokens ([{}]) can be still open on the right side,
                assuming that they will be closed in the following lines
            optional comment


        single line examples where match or case is promoted to a keyword:
            match x:  # random comment
            match (x):
            match(
            case []:
            case {"x": x,
            case {"x": x,  # another comment
            case (0, 0):

        single line examples where match or case is NOT promoted to a keyword:
            match x
            case [(]:
            case {"x": x,}
            case {"x": x,)
            case (0, 0)

        False positives, i.e. promoting an identifier to a keyword erroneously, are very
        unlikely. One example is the line "match(" because it could be either followed by
        line "x)" and make it a normal function call, or it could be followed by line
        "x):" and make it a match statement.
        """

        indMatchCase = None
        closingParens = {"(": ")", "[": "]", "{": "}"}
        parensStack = []
        for i, token in enumerate(tokens):
            if i == 0 and isinstance(token, NonIdentifierToken):
                if str(token).isspace():
                    continue  # ignore whitespace before match or case identifiers
                else:
                    return
            elif indMatchCase is None:
                if isinstance(token, IdentifierToken) and str(token) in (
                    "match",
                    "case",
                ):
                    indMatchCase = i
                else:
                    return
            elif isinstance(token, OpenParenToken):
                parensStack.append(closingParens[str(token)])
            elif isinstance(token, CloseParenToken):
                if len(parensStack) > 0 and parensStack[-1] == str(token):
                    parensStack.pop()
                else:
                    return  # invalid parentheses (or square brackets or curly brackets)

        if indMatchCase is None:
            return
        i2 = -1  # index of the last token, not counting comment tokens
        if isinstance(tokens[i2], CommentToken):
            i2 = -2  # ignore the comment token at the end
            if len(tokens) < 3:
                return

        if len(parensStack) > 0:
            pass  # probably a match or case keyword, depending on the lines that follow
        elif (
            isinstance(tokens[i2], NonIdentifierToken)
            and str(tokens[i2]).strip() == ":"
        ):
            pass  # very likely a match or case keyword
        else:
            return  # not a match or case keyword

        t = tokens[indMatchCase]
        tokens[indMatchCase] = KeywordToken(t.line, t.start, t.end)

    def _findEndOfString(self, line, token):
        """Find the end of a string. Returns (token, endToken). The first
        is the given token or a replacement (UnterminatedStringToken).
        The latter is None, or the BlockState. If given, the line is
        finished.

        """

        # Set state
        self._identifierState(None)

        # Find the matching end in the rest of the line
        # Do not use the start parameter of search, since ^ does not work then
        style = token._style
        endMatch = endProgs[style].search(line[token.end :])

        if endMatch:
            # The string does end on this line
            tokenArgs = line, token.start, token.end + endMatch.end()
            if style in ['"""', "'''"]:
                token = MultilineStringToken(*tokenArgs)
            else:
                token.end = token.end + endMatch.end()
            return [token]
        else:
            # The string does not end on this line
            tokenArgs = line, token.start, token.end + len(line)
            if style == "'''":
                return [MultilineStringToken(*tokenArgs), BlockState(1)]
            elif style == '"""':
                return [MultilineStringToken(*tokenArgs), BlockState(2)]
            else:
                lineContMatch = stringLineContinuation.search(line[token.end :])
                if lineContMatch:
                    return [
                        StringToken(*tokenArgs),
                        BlockState(5 if style == "'" else 6),
                    ]
                return [UnterminatedStringToken(*tokenArgs)]

    def _findNextToken(self, line, pos):
        """Returns a token or None if no new tokens can be found."""

        # Init tokens, if pos too large, we are done
        if pos > len(line):
            return None
        tokens = []

        # Find the start of the next string or comment
        match = tokenProg.search(line, pos)

        # Process the Non-Identifier between pos and match.start()
        # or end of line
        nonIdentifierEnd = match.start() if match else len(line)

        # Return the Non-Identifier token if non-null
        # todo: here it goes wrong (allow returning more than one token?)
        token = NonIdentifierToken(line, pos, nonIdentifierEnd)
        strippedNonIdentifier = str(token).strip()
        if token:
            tokens.append(token)

        # Do checks for line continuation and identifierState
        # Is the last non-whitespace a line-continuation character?
        if strippedNonIdentifier.endswith("\\"):
            lineContinuation = True
            # If there are non-whitespace characters after def or class,
            # cancel the identifierState
            if strippedNonIdentifier != "\\":
                self._identifierState(None)
        else:
            lineContinuation = False
            # If there are non-whitespace characters after def or class,
            # cancel the identifierState
            if strippedNonIdentifier != "":
                self._identifierState(None)

        # If no match, we are done processing the line
        if not match:
            if lineContinuation:
                tokens.append(BlockState(self._identifierState()))
            return tokens

        # The rest is to establish what identifier we are dealing with

        # Comment
        if match.group() == "#":
            matchStart = match.start()
            if not line[:matchStart].strip() and (
                line[matchStart:].startswith(("##", "#%%", "# %%"))
            ):
                tokens.append(CellCommentToken(line, matchStart, len(line)))
            elif self._isTodoItem(line[matchStart + 1 :]):
                tokens.append(TodoCommentToken(line, matchStart, len(line)))
            else:
                tokens.append(CommentToken(line, matchStart, len(line)))
            if lineContinuation:
                tokens.append(BlockState(self._identifierState()))
            return tokens

        # If there are non-whitespace characters after def or class,
        # cancel the identifierState (this time, also if there is just a \
        # since apparently it was not on the end of a line)
        if strippedNonIdentifier != "":
            self._identifierState(None)

        # Identifier ("a word or number") Find out whether it is a key word
        if match.group(4) is not None:
            identifier = match.group(4)
            tokenArgs = line, match.start(), match.end()

            # Set identifier state
            identifierState = self._identifierState(identifier)

            if identifier in self._keywords:
                tokens.append(KeywordToken(*tokenArgs))
            elif identifier in self._builtins and (
                "." + identifier not in line and "def " + identifier not in line
            ):
                tokens.append(BuiltinsToken(*tokenArgs))
            elif identifier in self._instance:
                tokens.append(InstanceToken(*tokenArgs))
            elif identifier[0] in "0123456789":
                self._identifierState(None)
                tokens.append(NumberToken(*tokenArgs))
            else:
                if identifierState == 3 and line[match.end() :].lstrip().startswith(
                    "("
                ):
                    tokens.append(FunctionNameToken(*tokenArgs))
                elif identifierState == 4:
                    tokens.append(ClassNameToken(*tokenArgs))
                else:
                    tokens.append(IdentifierToken(*tokenArgs))

        elif match.group(3) is not None:
            # We have matched a string-start
            # Find the string style ( ' or " or ''' or """)
            token = StringToken(line, match.start(), match.end())
            token._style = match.group(3)  # The style is in match group 3
            tokens.append(token)
        elif match.group(5) is not None:
            token = OpenParenToken(line, match.start(), match.end())
            token._style = match.group(5)
            tokens.append(token)
        elif match.group(6) is not None:
            token = CloseParenToken(line, match.start(), match.end())
            token._style = match.group(6)
            tokens.append(token)
        elif match.group(7) is not None:
            token = IllegalToken(line, match.start(), match.end())
            token._style = match.group(7)
            tokens.append(token)
        # Done
        return tokens


class PythonParser(PythonParser):  # Ambiguous Python parser
    """Parser for either Python2 or Python3, and we do not know which."""

    _extensions = [".py", ".pyw"]
    _shebangKeywords = ["python"]
    # The list of keywords is overridden by the Python2/3 specific parsers
    _keywords = pythonKeywords
    # The list of builtins and instances is overridden by the Python2/3 specific parsers
    _builtins = pythonBuiltins
    _instance = pythonInstance

    @classmethod
    def disambiguate(cls, text):
        # try to look into the source...
        # or ... well, ppl should use Python3. Use a shebang to annotate a Python file as Python 2
        return "python3"


class Python2Parser(PythonParser):
    """Parser for Python 2.x code."""

    # The application should choose whether to set the Py 2 specific parser
    _extensions = []
    _shebangKeywords = ["python2"]
    _keywords = python2Keywords
    _builtins = python2Builtins
    _instance = python2Instance


class Python3Parser(PythonParser):
    """Parser for Python 3.x code."""

    # The application should choose whether to set the Py 3 specific parser
    _extensions = []
    _shebangKeywords = ["python3"]
    _keywords = python3Keywords
    _builtins = python3Builtins
    _instance = python3Instance



# -*- coding: utf-8 -*-
# This example demonstrates simple req rep.
# This example can be run in one go or in two parts running in
# different processes.

## ========== One end

import yoton

verbosity = 0


# Create a replier class by subclassing RepChannel
class Adder(yoton.RepChannel):
    def add(self, item1, item2):
        return item1 + item2


# Create a context and a rep channel
ct1 = yoton.Context(verbose=verbosity)
rep = Adder(ct1, "duplicate")

# Connect and turn duplicator on
ct1.bind("localhost:test")
rep.set_mode("thread")


## ========== Other end

import yoton

verbosity = 0

# Create a context and a req channel
ct2 = yoton.Context(verbose=verbosity)
req = yoton.ReqChannel(ct2, "duplicate")

# Connect
ct2.connect("localhost:test")

# Duplicate a string
print(req.add("foo", "bar").result(1))
print(req.add(3, 4).result(1))



"""
Code editor extensions that change its appearance
"""

from ..qt import QtGui, QtCore, QtWidgets

Qt = QtCore.Qt

from ..misc import ce_option
from ..manager import Manager

# todo: what about calling all extensions. CE_HighlightCurrentLine,
# or EXT_HighlightcurrentLine?

from ..parsers.tokens import ParenthesisToken
import enum


class HighlightMatchingOccurrences:
    # Register style element
    _styleElements = [
        (
            "Editor.Highlight matching occurrences",
            "The background color to highlight matching occurrences of the currently selected word.",
            "back:#fdfda3",
        )
    ]

    def highlightMatchingOccurrences(self):
        """highlightMatchingOccurrences()

        Get whether to highlight matching occurrences.
        """
        return self.__highlightMatchingOccurrences

    @ce_option(True)
    def setHighlightMatchingOccurrences(self, value):
        """setHighlightMatchingOccurrences(value)

        Set whether to highlight matching occurrences.
        """
        self.__highlightMatchingOccurrences = bool(value)
        self.viewport().update()

    def _doHighlight(self, text):
        if text.strip() != text or "\u2029" in text:
            # selection has leading/trailing whitespace or contains a line break
            return

        # make cursor at the beginning of the first visible block
        cursor = self.cursorForPosition(QtCore.QPoint(0, 0))
        cursor.movePosition(cursor.MoveOperation.StartOfWord)
        doc = self.document()

        color = self.getStyleElementFormat("editor.highlightMatchingOccurrences").back
        painter = QtGui.QPainter()
        painter.begin(self.viewport())
        painter.setBrush(color)
        painter.setPen(color.darker(110))

        # flag "FindWholeWords" in doc.find would not consider "_" as a word character
        # therefore use a custom regular expression instead
        QRE = QtCore.QRegularExpression
        needle = QRE(r"\b" + QRE.escape(text) + r"\b")

        # find occurrences
        for i in range(500):
            cursor = doc.find(needle, cursor, doc.FindFlag.FindCaseSensitively)
            if cursor is None or cursor.isNull():
                # no more matches
                break

            # don't highlight the actual selection
            if cursor == self.textCursor():
                continue

            endRect = self.cursorRect(cursor)
            cursor.setPosition(min(cursor.position(), cursor.anchor()))
            startRect = self.cursorRect(cursor)

            if startRect.top() > self.height():
                # rest of document is not visible, don't bother highlighting
                break

            width = endRect.left() - startRect.left()
            cursorHeight = startRect.height()

            heightDiff = endRect.top() - startRect.top()
            if heightDiff == 0:
                painter.drawRect(startRect.left(), startRect.top(), width, cursorHeight)
            elif heightDiff > 0:
                cursor.movePosition(cursor.MoveOperation.EndOfLine)
                secondLineY = self.cursorRect(cursor).top()
                endLineY = endRect.top()
                fullLineStartX = 0
                fullLineEndX = self.width()

                # first partial line
                width = fullLineEndX - startRect.left()
                painter.drawRect(startRect.left(), startRect.top(), width, cursorHeight)

                # full lines in between
                if endLineY > secondLineY:
                    width = fullLineEndX - fullLineStartX
                    height = endLineY - secondLineY
                    painter.drawRect(fullLineStartX, secondLineY, width, height)

                # last partial line
                width = endRect.left() - fullLineStartX
                if width > 0:
                    painter.drawRect(fullLineStartX, endLineY, width, cursorHeight)

            # move to end of word again, otherwise we never advance in the doc
            cursor.movePosition(cursor.MoveOperation.EndOfWord)
        else:
            print("Matching selection highlighting did not break")

        painter.end()

    def paintEvent(self, event):
        """Paints behinds its super()."""
        cursor = self.textCursor()
        if self.__highlightMatchingOccurrences and cursor.hasSelection():
            text = cursor.selectedText()
            self._doHighlight(text)

        super().paintEvent(event)


class _ParenNotFound(Exception):
    pass


class _ParenIterator:
    """Iterates in given direction over parentheses in the document.
    Uses the stored token-list of the blocks.
    Iteration gives both a parenthesis and its global position."""

    def __init__(self, cursor, direction):
        self.cur_block = cursor.block()
        self.cur_tokens = self._getParenTokens()
        self.direction = direction
        # We need to know where we start in the current token list
        k = 0
        try:
            while self.cur_tokens[k].end != cursor.positionInBlock():
                k += 1
            self.cur_pos = k
        except IndexError:
            # If the parenthesis cannot be found, it means that it is not included
            # in any token, ie. it is part of a string or comment
            raise _ParenNotFound

    def _getParenTokens(self):
        try:
            return [
                x
                for x in self.cur_block.userData().tokens
                if isinstance(x, ParenthesisToken)
            ]
        except AttributeError:
            return []  # can be a piece of text that we do not tokenize or have not stored tokens

    def __iter__(self):
        return self

    def __next__(self):
        self.cur_pos += self.direction
        while self.cur_pos >= len(self.cur_tokens) or self.cur_pos < 0:
            if self.direction == 1:
                self.cur_block = self.cur_block.next()
            else:
                self.cur_block = self.cur_block.previous()
            if not self.cur_block.isValid():
                raise StopIteration
            self.cur_tokens = self._getParenTokens()
            if self.direction == 1:
                self.cur_pos = 0
            else:
                self.cur_pos = len(self.cur_tokens) - 1
        return (
            self.cur_tokens[self.cur_pos]._style,
            self.cur_block.position() + self.cur_tokens[self.cur_pos].end,
        )


class _PlainTextParenIterator:
    """Iterates in given direction over parentheses in the document.
    To be used when there is no parser.
    Iteration gives both a parenthesis and its global position."""

    def __init__(self, cursor, direction):
        self.fulltext = cursor.document().toPlainText()
        self.position = cursor.position() - 1
        self.direction = direction

    def __iter__(self):
        return self

    def __next__(self):
        self.position += self.direction
        try:
            while self.fulltext[self.position] not in "([{)]}":
                self.position += self.direction
                if self.position < 0:
                    raise StopIteration
        except IndexError:
            raise StopIteration
        return self.fulltext[self.position], self.position + 1


class _MatchStatus(enum.Enum):
    NoMatch = 0
    Match = 1
    MisMatch = 2


class _MatchResult:
    def __init__(self, status, corresponding=None, offending=None):
        self.status = status
        self.corresponding = corresponding
        self.offending = offending


class HighlightMatchingBracket:
    # Register style element
    _styleElements = [
        (
            "Editor.Highlight matching bracket",
            "The background color to highlight matching brackets.",
            "back:#ccc",
        ),
        (
            "Editor.Highlight unmatched bracket",
            "The background color to highlight unmatched brackets.",
            "back:#F7BE81",
        ),
        (
            "Editor.Highlight mismatching bracket",
            "The background color to highlight mismatching brackets.",
            "back:#F7819F",
        ),
    ]

    _BRACKS_OPEN = "([{"
    _BRACKS_CLOSE = ")]}"
    _BRACKS = _BRACKS_OPEN + _BRACKS_CLOSE
    _matchingBrackets = dict(
        zip(_BRACKS_OPEN + _BRACKS_CLOSE, _BRACKS_CLOSE + _BRACKS_OPEN)  # noqa: B905
    )

    def highlightMatchingBracket(self):
        """Get whether to highlight matching brackets."""
        return self.__highlightMatchingBracket

    @ce_option(True)
    def setHighlightMatchingBracket(self, value):
        """Set whether to highlight matching brackets."""
        self.__highlightMatchingBracket = bool(value)
        self.viewport().update()

    def highlightMisMatchingBracket(self):
        """Get whether to highlight mismatching brackets."""
        return self.__highlightMisMatchingBracket

    @ce_option(True)
    def setHighlightMisMatchingBracket(self, value):
        """Set whether to highlight mismatching brackets."""
        self.__highlightMisMatchingBracket = bool(value)
        self.viewport().update()

    def _highlightSingleChar(self, painter, cursor, width, colorname):
        """Draws a highlighting rectangle around the single character to the
        left of the specified cursor.
        """
        cursor_rect = self.cursorRect(cursor)
        top = cursor_rect.top()
        left = cursor_rect.left() - width
        height = cursor_rect.bottom() - top + 1
        color = self.getStyleElementFormat(colorname).back
        painter.setBrush(color)
        painter.setPen(color.darker(110))
        painter.drawRect(QtCore.QRect(int(left), int(top), int(width), int(height)))

    def _findMatchingBracket(self, char, cursor):
        """Find a bracket that matches the specified char in the specified document.
        Return a _MatchResult object indicating whether this succeded and the
        positions of the parentheses causing this result.
        """
        if char in self._BRACKS_CLOSE:
            direction = -1
            stacking = self._BRACKS_CLOSE
            unstacking = self._BRACKS_OPEN
        elif char in self._BRACKS_OPEN:
            direction = 1
            stacking = self._BRACKS_OPEN
            unstacking = self._BRACKS_CLOSE
        else:
            raise ValueError("invalid bracket character: " + char)

        stacked_paren = [(char, cursor.position())]  # using a Python list as a stack
        # stack not empty because the _ParenIterator will not give back
        # the parenthesis we're matching
        our_iterator = (
            _ParenIterator
            if self.parser() is not None and self.parser().name() != ""
            else _PlainTextParenIterator
        )
        for paren, pos in our_iterator(cursor, direction):
            if paren in stacking:
                stacked_paren.append((paren, pos))
            elif paren in unstacking:
                if self._matchingBrackets[stacked_paren[-1][0]] != paren:
                    return _MatchResult(
                        _MatchStatus.MisMatch, pos, stacked_paren[-1][1]
                    )
                else:
                    stacked_paren.pop()

            if len(stacked_paren) == 0:
                # we've found our match
                return _MatchResult(_MatchStatus.Match, pos)
        return _MatchResult(_MatchStatus.NoMatch)

    def _cursorAt(self, doc, pos):
        new_cursor = QtGui.QTextCursor(doc)
        new_cursor.setPosition(pos)
        return new_cursor

    def paintEvent(self, event):
        """Paints behinds its super().

        If the current cursor is positioned to the right of a bracket ()[]{},
        look for a matching one, and, if found, draw a highlighting rectangle
        around both brackets of the pair.
        """
        if not self.__highlightMatchingBracket:
            super().paintEvent(event)
            return

        cursor = QtGui.QTextCursor(self.textCursor())

        # We need to clear the selection because otherwise movePosition(MoveOperation.Right)
        # moves the cursor to the right end of the selection instead of advancing by one step.
        cursor.clearSelection()

        if cursor.atBlockStart():
            cursor.movePosition(cursor.MoveOperation.Right)
            movedRight = True
        else:
            movedRight = False
        text = cursor.block().text()
        pos = cursor.positionInBlock() - 1

        if len(text) > pos and len(text) > 0:
            # get the character to the left of the cursor
            char = text[pos]

            if not movedRight and char not in self._BRACKS and len(text) > pos + 1:
                # no brace to the left of cursor; try to the right
                cursor.movePosition(cursor.MoveOperation.Right)
                char = text[pos + 1]

            if char in self._BRACKS:
                doc = cursor.document()
                try:
                    match_res = self._findMatchingBracket(char, cursor)
                    fm = QtGui.QFontMetrics(doc.defaultFont())
                    width = fm.horizontalAdvance(
                        char
                    )  # assumes that both paren have the same width
                    painter = QtGui.QPainter()
                    painter.begin(self.viewport())
                    if match_res.status == _MatchStatus.NoMatch:
                        self._highlightSingleChar(
                            painter, cursor, width, "editor.highlightUnmatchedBracket"
                        )
                    elif match_res.status == _MatchStatus.Match:
                        self._highlightSingleChar(
                            painter, cursor, width, "editor.highlightMatchingBracket"
                        )
                        self._highlightSingleChar(
                            painter,
                            self._cursorAt(doc, match_res.corresponding),
                            width,
                            "editor.highlightMatchingBracket",
                        )
                    else:  # this is a mismatch
                        if (
                            cursor.position() != match_res.offending
                            or not self.highlightMisMatchingBracket()
                        ):
                            self._highlightSingleChar(
                                painter,
                                cursor,
                                width,
                                "editor.highlightUnmatchedBracket",
                            )
                        if self.highlightMisMatchingBracket():
                            self._highlightSingleChar(
                                painter,
                                self._cursorAt(doc, match_res.corresponding),
                                width,
                                "editor.highlightMisMatchingBracket",
                            )
                            self._highlightSingleChar(
                                painter,
                                self._cursorAt(doc, match_res.offending),
                                width,
                                "editor.highlightMisMatchingBracket",
                            )

                    painter.end()
                except _ParenNotFound:
                    # is raised when current parenthesis is not
                    # found in its line token list, meaning it is in a string literal
                    pass

        super().paintEvent(event)


class HighlightCurrentLine:
    """
    Highlight the current line
    """

    # Register style element
    _styleElements = [
        (
            "Editor.Highlight current line",
            "The background color of the current line highlight.",
            "back:#ffff99",
        )
    ]

    def highlightCurrentLine(self):
        """Get whether to highlight the current line."""
        return self.__highlightCurrentLine

    @ce_option(True)
    def setHighlightCurrentLine(self, value):
        """Set whether to highlight the current line."""
        self.__highlightCurrentLine = bool(value)
        self.viewport().update()

    def paintEvent(self, event):
        """Paints behind its super()

        Paints a rectangle spanning the current block (in case of line wrapping, this
        means multiple lines)
        """
        if not self.highlightCurrentLine():
            super().paintEvent(event)
            return

        # Get color
        color = self.getStyleElementFormat("editor.highlightCurrentLine").back

        # Find the top of the current block, and the height
        cursor = self.textCursor()
        cursor.movePosition(cursor.MoveOperation.StartOfBlock)
        top = self.cursorRect(cursor).top()
        cursor.movePosition(cursor.MoveOperation.EndOfBlock)
        height = self.cursorRect(cursor).bottom() - top + 1

        margin = self.document().documentMargin()
        painter = QtGui.QPainter()
        painter.begin(self.viewport())
        painter.fillRect(
            QtCore.QRect(
                int(margin),
                int(top),
                int(self.viewport().width() - 2 * margin),
                int(height),
            ),
            color,
        )
        painter.end()

        super().paintEvent(event)

        # for debugging paint events
        # if 'log' not in self.__class__.__name__.lower():
        #    print(height, event.rect().width())


class IndentationGuides:
    # Register style element
    _styleElements = [
        (
            "Editor.Indentation guides",
            "The color and style of the indentation guides.",
            "fore:#DDF,linestyle:solid",
        )
    ]

    def showIndentationGuides(self):
        """Get whether to show indentation guides."""
        return self.__showIndentationGuides

    @ce_option(True)
    def setShowIndentationGuides(self, value):
        """Set whether to show indentation guides."""
        self.__showIndentationGuides = bool(value)
        self.viewport().update()

    def paintEvent(self, event):
        """Paint the indentation guides, using the indentation info calculated
        by the highlighter.
        """
        super().paintEvent(event)

        if not self.showIndentationGuides():
            return

        # Get doc and viewport
        doc = self.document()
        viewport = self.viewport()

        # Get multiplication factor and indent width
        indentWidth = self.indentWidth()
        if self.indentUsingSpaces():
            factor = 1
        else:
            factor = indentWidth

        # Init painter
        painter = QtGui.QPainter()
        painter.begin(viewport)

        # Prepare pen
        format = self.getStyleElementFormat("editor.IndentationGuides")
        pen = QtGui.QPen(format.fore)
        pen.setStyle(format.linestyle)
        painter.setPen(pen)
        offset = doc.documentMargin() + self.contentOffset().x()

        def paintIndentationGuides(cursor):
            y3 = self.cursorRect(cursor).top()
            y4 = self.cursorRect(cursor).bottom()

            bd = cursor.block().userData()
            if bd and hasattr(bd, "indentation") and bd.indentation:
                for x in range(indentWidth, bd.indentation * factor, indentWidth):
                    w = self.fontMetrics().horizontalAdvance("i" * x) + offset
                    w += 1  # Put it more under the block
                    if w > 0:  # if scrolled horizontally it can become < 0
                        painter.drawLine(QtCore.QLine(int(w), int(y3), int(w), int(y4)))

        self.doForVisibleBlocks(paintIndentationGuides)

        # Done
        painter.end()


class FullUnderlines:
    def paintEvent(self, event):
        """Paint a horizontal line for the blocks for which there is a
        syntax format that has underline:full. Whether this is the case
        is stored at the blocks user data.
        """
        super().paintEvent(event)

        painter = QtGui.QPainter()
        painter.begin(self.viewport())

        margin = self.document().documentMargin()
        w = self.viewport().width()

        def paintUnderline(cursor):
            y = self.cursorRect(cursor).bottom()
            fullUnderlineFormat = getattr(
                cursor.block().userData(), "fullUnderlineFormat", None
            )
            if fullUnderlineFormat is not None:
                # Apply pen
                pen = QtGui.QPen(fullUnderlineFormat.fore)
                pen.setStyle(fullUnderlineFormat.linestyle)
                painter.setPen(pen)
                # Paint
                painter.drawLine(
                    QtCore.QLine(int(margin), int(y), int(w - 2 * margin), int(y))
                )

        self.doForVisibleBlocks(paintUnderline)

        painter.end()


class CodeFolding:
    def paintEvent(self, event):
        super().paintEvent(event)

        return  # Code folding code is not yet complete

        painter = QtGui.QPainter()
        painter.begin(self.viewport())

        margin = int(self.document().documentMargin())

        def paintCodeFolders(cursor):
            y = int(self.cursorRect(cursor).top())
            h = int(self.cursorRect(cursor).height())
            rect = QtCore.QRect(margin, y, h, h)
            text = cursor.block().text()
            if text.rstrip().endswith(":"):
                painter.drawRect(rect)
                AF = QtCore.Qt.AlignmentFlag
                painter.drawText(rect, AF.AlignVCenter | AF.AlignHCenter, "-")
                # Apply pen

                # Paint
                # painter.drawLine(QtCore.QLine(int(margin), int(y), int(w - 2*margin), int(y)))

        self.doForVisibleBlocks(paintCodeFolders)

        painter.end()


class LongLineIndicator:
    # Register style element
    _styleElements = [
        (
            "Editor.Long line indicator",
            "The color and style of the long line indicator.",
            "fore:#BBB,linestyle:solid",
        )
    ]

    def longLineIndicatorPosition(self):
        """Get the position of the long line indicator (aka edge column).
        A value of 0 or smaller means that no indicator is shown.
        """
        return self.__longLineIndicatorPosition

    @ce_option(80)
    def setLongLineIndicatorPosition(self, value):
        """Set the position of the long line indicator (aka edge column).
        A value of 0 or smaller means that no indicator is shown.
        """
        self.__longLineIndicatorPosition = int(value)
        self.viewport().update()

    def paintEvent(self, event):
        """Paint the long line indicator. Paints behind its super()"""
        if self.longLineIndicatorPosition() <= 0:
            super().paintEvent(event)
            return

        # Get doc and viewport
        doc = self.document()
        viewport = self.viewport()

        # Get position of long line
        fm = self.fontMetrics()
        # horizontalAdvance of ('i'*length) not length * (horizontalAdvance of 'i') b/c of
        # font kerning and rounding
        x = fm.horizontalAdvance("i" * self.longLineIndicatorPosition())
        x += doc.documentMargin() + self.contentOffset().x()
        x += 1  # Move it a little next to the cursor

        # Prepare painter
        painter = QtGui.QPainter()
        painter.begin(viewport)

        # Prepare pen
        format = self.getStyleElementFormat("editor.LongLineIndicator")
        pen = QtGui.QPen(format.fore)
        pen.setStyle(format.linestyle)
        painter.setPen(pen)

        # Draw line and end painter
        painter.drawLine(QtCore.QLine(int(x), 0, int(x), int(viewport.height())))
        painter.end()

        # Propagate event
        super().paintEvent(event)


class ShowWhitespace:
    def showWhitespace(self):
        """Show or hide whitespace markers"""
        option = self.document().defaultTextOption()
        return bool(option.flags() & option.ShowTabsAndSpaces)

    @ce_option(False)
    def setShowWhitespace(self, value):
        try:
            option = self.document().defaultTextOption()
            if value:
                option.setFlags(option.flags() | option.Flag.ShowTabsAndSpaces)
            else:
                option.setFlags(option.flags() & ~option.Flag.ShowTabsAndSpaces)
            self.document().setDefaultTextOption(option)
        except Exception:
            # This can produce: 2147483617 is not a valid QTextOption.Flag
            # and I do not know how to avoid it :/
            pass


class ShowLineEndings:
    @ce_option(False)
    def showLineEndings(self):
        """Get whether line ending markers are shown."""
        option = self.document().defaultTextOption()
        return bool(option.flags() & option.ShowLineAndParagraphSeparators)

    def setShowLineEndings(self, value):
        try:
            option = self.document().defaultTextOption()
            if value:
                option.setFlags(
                    option.flags() | option.Flag.ShowLineAndParagraphSeparators
                )
            else:
                option.setFlags(
                    option.flags() & ~option.Flag.ShowLineAndParagraphSeparators
                )
            self.document().setDefaultTextOption(option)
        except Exception:
            # This can produce: 2147483617 is not a valid QTextOption.Flag
            # and I do not know how to avoid it :/
            pass


class LineNumbers:
    # Margin on both side of the line numbers
    _LineNumberAreaMargin = 3

    # Register style element
    _styleElements = [
        (
            "Editor.Line numbers",
            "The text- and background-color of the line numbers.",
            "fore:#222,back:#DDD",
        )
    ]

    class __LineNumberArea(QtWidgets.QWidget):
        """This is the widget reponsible for drawing the line numbers."""

        def __init__(self, codeEditor):
            super().__init__(codeEditor)
            self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
            self._pressedY = None
            self._lineNrChoser = None

        def _getY(self, pos):
            tmp = self.mapToGlobal(pos)
            return self.parent().viewport().mapFromGlobal(tmp).y()

        def mousePressEvent(self, event):
            self._pressedY = self._getY(event.position().toPoint())

        def mouseReleaseEvent(self, event):
            self._handleWholeBlockSelection(self._getY(event.position().toPoint()))

        def mouseMoveEvent(self, event):
            self._handleWholeBlockSelection(self._getY(event.position().toPoint()))

        def _handleWholeBlockSelection(self, y2):
            # Get y1 and sort (y1, y2)
            y1 = self._pressedY
            if y1 is None:
                y1 = y2
            if y2 < y1:
                y1, y2 = y2, y1

            # Get cursor and two cursors corresponding to selected blocks
            editor = self.parent()
            cursor = editor.textCursor()
            c1 = editor.cursorForPosition(QtCore.QPoint(0, int(y1)))
            c2 = editor.cursorForPosition(QtCore.QPoint(0, int(y2)))

            # Make these two cursors select the whole block
            c1.movePosition(c1.MoveOperation.StartOfBlock, c1.MoveMode.MoveAnchor)
            c2.movePosition(c2.MoveOperation.EndOfBlock, c2.MoveMode.MoveAnchor)

            # Apply selection
            cursor.setPosition(c1.position(), cursor.MoveMode.MoveAnchor)
            cursor.setPosition(c2.position(), cursor.MoveMode.KeepAnchor)
            editor.setTextCursor(cursor)

        def mouseDoubleClickEvent(self, event):
            self.showLineNumberChoser()

        def showLineNumberChoser(self):
            # Create line number choser if needed
            if self._lineNrChoser is None:
                self._lineNrChoser = LineNumbers.LineNumberChoser(self.parent())
            # Get editor and cursor
            editor = self.parent()
            cursor = editor.textCursor()
            # Get (x,y) pos and apply
            x, y = self.width() + 4, editor.cursorRect(cursor).y()
            self._lineNrChoser.move(int(x), int(y))
            # Show/reset line number choser
            self._lineNrChoser.reset(cursor.blockNumber() + 1)

        def paintEvent(self, event):
            editor = self.parent()

            if not editor.showLineNumbers():
                return

            # Get doc and viewport
            viewport = editor.viewport()

            # Get format and margin
            format = editor.getStyleElementFormat("editor.LineNumbers")
            margin = editor._LineNumberAreaMargin

            # Init painter
            painter = QtGui.QPainter()
            painter.begin(self)

            # Get which part to paint. Just do all to avoid glitches
            w = editor._getLineNumberAreaWidth()
            y1, y2 = 0, editor.height()
            # y1, y2 = event.rect().top()-10, event.rect().bottom()+10

            # Get offset
            tmp = self.mapToGlobal(QtCore.QPoint(0, 0))
            offset = viewport.mapFromGlobal(tmp).y()

            # Draw the background
            painter.fillRect(QtCore.QRect(0, int(y1), int(w), int(y2)), format.back)

            # Get cursor
            cursor = editor.cursorForPosition(QtCore.QPoint(0, int(y1)))

            # Prepare fonts
            font1 = editor.font()
            font2 = editor.font()
            font2.setBold(True)
            currentBlockNumber = editor.textCursor().block().blockNumber()

            # Init painter with font and color
            painter.setFont(font1)
            painter.setPen(format.fore)

            # Repainting always starts at the first block in the viewport,
            # regardless of the event.rect().y(). Just to keep it simple
            while True:
                blockNumber = cursor.block().blockNumber()

                y = editor.cursorRect(cursor).y()

                # Set font to bold if line number is the current
                if blockNumber == currentBlockNumber:
                    painter.setFont(font2)

                AF = Qt.AlignmentFlag
                painter.drawText(
                    0, y - offset, w - margin, 50, AF.AlignRight, str(blockNumber + 1)
                )

                # Set font back
                if blockNumber == currentBlockNumber:
                    painter.setFont(font1)

                if y > y2:
                    break  # Reached end of the repaint area
                if not cursor.block().next().isValid():
                    break  # Reached end of the text

                cursor.movePosition(cursor.MoveOperation.NextBlock)

            # Done
            painter.end()

    class LineNumberChoser(QtWidgets.QSpinBox):
        def __init__(self, parent):
            super().__init__(parent)
            self._editor = parent

            ss = (
                "QSpinBox { border: 2px solid #789; border-radius: 3px; padding: 4px; }"
            )
            self.setStyleSheet(ss)

            self.setPrefix("Go to line: ")
            self.setAccelerated(True)
            self.setButtonSymbols(self.ButtonSymbols.NoButtons)
            self.setCorrectionMode(self.CorrectionMode.CorrectToNearestValue)

            # Signal for when value changes, and flag to disable it once
            self._ignoreSignalOnceFlag = False
            self.valueChanged.connect(self.onValueChanged)

        def reset(self, currentLineNumber):
            # Set value to (given) current line number
            self._ignoreSignalOnceFlag = True
            self.setRange(1, self._editor.blockCount())
            self.setValue(currentLineNumber)
            # Select text and focus so that the user can simply start typing
            self.selectAll()
            self.setFocus()
            # Make visible
            self.show()
            self.raise_()

        def focusOutEvent(self, event):
            self.hide()

        def keyPressEvent(self, event):
            if event.key() in [
                QtCore.Qt.Key.Key_Escape,
                QtCore.Qt.Key.Key_Enter,
                QtCore.Qt.Key.Key_Return,
            ]:
                self._editor.setFocus()  # Moves focus away, thus hiding self
            else:
                super().keyPressEvent(event)

        def onValueChanged(self, nr):
            if self._ignoreSignalOnceFlag:
                self._ignoreSignalOnceFlag = False
            else:
                self._editor.gotoLine(nr)

    def __init__(self, *args, **kwds):
        self.__lineNumberArea = None
        self.__leftMarginHandle = None
        super().__init__(*args, **kwds)
        # Create widget that draws the line numbers
        self.__lineNumberArea = self.__LineNumberArea(self)
        # Issue an update when the font or amount of line numbers changes
        self.blockCountChanged.connect(self.__onBlockCountChanged)
        self.fontChanged.connect(self.__onBlockCountChanged)
        self.__leftMarginHandle = self._setLeftBarMargin(
            self.__leftMarginHandle, self._getLineNumberAreaWidth()
        )
        self.__onBlockCountChanged()

    def gotoLinePopup(self):
        """Popup the little widget to quickly goto a certain line.
        Can also be achieved by double-clicking the line number area.
        """
        self.__lineNumberArea.showLineNumberChoser()

    def showLineNumbers(self):
        return self.__showLineNumbers

    @ce_option(True)
    def setShowLineNumbers(self, value):
        self.__showLineNumbers = bool(value)
        # Note that this method is called before the __init__ is finished,
        # so that the __lineNumberArea is not yet created.
        if self.__lineNumberArea:
            if self.__showLineNumbers:
                self.__lineNumberArea.show()
            else:
                self.__lineNumberArea.hide()
            if self.__leftMarginHandle is not None:
                self._setLeftBarMargin(
                    self.__leftMarginHandle, self._getLineNumberAreaWidth()
                )

    def _getLineNumberAreaWidth(self):
        """Count the number of lines, compute the length of the longest line number
        (in pixels)
        """
        if not self.__showLineNumbers:
            return 0
        lastLineNumber = self.blockCount()
        margin = self._LineNumberAreaMargin
        return self.fontMetrics().horizontalAdvance(str(lastLineNumber)) + 2 * margin

    def __onBlockCountChanged(self, count=None):
        """Update the line number area width. This requires to set the
        viewport margins, so there is space to draw the linenumber area
        """
        if self.__showLineNumbers:
            self.__leftMarginHandle = self._setLeftBarMargin(
                self.__leftMarginHandle, self._getLineNumberAreaWidth()
            )

    def resizeEvent(self, event):
        super().resizeEvent(event)

        # On resize, resize the lineNumberArea, too
        rect = self.contentsRect()
        m = self._getMarginBeforeLeftBar(self.__leftMarginHandle)
        w = self._getLineNumberAreaWidth()
        self.__lineNumberArea.setGeometry(rect.x() + m, rect.y(), w, rect.height())

    def paintEvent(self, event):
        super().paintEvent(event)
        # On repaint, update the complete line number area
        w = self._getLineNumberAreaWidth()
        self.__lineNumberArea.update(0, 0, w, self.height())


class BreakPoints:
    # Register style element
    _styleElements = [
        (
            "Editor.BreakPoints",
            "The fore- and background-color of the breakpoints.",
            "fore:#F66,back:#dfdfe1",
        )
    ]

    class __BreakPointArea(QtWidgets.QWidget):
        """This is the widget reponsible for drawing the break points."""

        def __init__(self, codeEditor):
            super().__init__(codeEditor)
            self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
            self.setMouseTracking(True)
            self._virtualBreakpoint = 0

        def _getY(self, pos):
            tmp = self.mapToGlobal(pos)
            return self.parent().viewport().mapFromGlobal(tmp).y()

        def mousePressEvent(self, event):
            self._toggleBreakPoint(self._getY(event.position().toPoint()))

        def mouseMoveEvent(self, event):
            y = self._getY(event.position().toPoint())
            editor = self.parent()
            c1 = editor.cursorForPosition(QtCore.QPoint(0, int(y)))
            self._virtualBreakpoint = c1.blockNumber() + 1
            self.update()

        def leaveEvent(self, event):
            self._virtualBreakpoint = 0
            self.update()

        def _toggleBreakPoint(self, y):
            # Get breakpoint corresponding to pressed pos
            editor = self.parent()
            c1 = editor.cursorForPosition(QtCore.QPoint(0, int(y)))
            linenr = c1.blockNumber() + 1
            # Toggle
            self.parent().toggleBreakpoint(linenr)

        def paintEvent(self, event):
            editor = self.parent()

            if not editor.showBreakPoints():
                return

            # Get format and margin
            format = editor.getStyleElementFormat("editor.breakpoints")
            w = editor._breakPointWidth
            marginX = max(1, int(0.1 * w))
            bulletWidth = w - 2 * marginX

            # Init painter
            painter = QtGui.QPainter()
            painter.begin(self)

            # Get which part to paint. Just do all to avoid glitches
            y1, y2 = 0, editor.height()

            # Draw the background
            painter.fillRect(QtCore.QRect(0, int(y1), int(w), int(y2)), format.back)

            # Get debug indicator and list of sorted breakpoints
            debugBlockIndicator = editor._debugLineIndicator - 1
            virtualBreakpoint = self._virtualBreakpoint - 1
            breakpoints = self.parent()._breakPoints
            if not (
                len(breakpoints) > 0
                or editor._debugLineIndicator
                or editor._debugLineIndicators
                or virtualBreakpoint > 0
            ):
                return

            # Get cursor
            cursor = editor.cursorForPosition(QtCore.QPoint(0, int(y1)))

            # Get start block number and bullet offset in pixels
            startBlockNumber = cursor.block().blockNumber()
            marginY = int(0.5 * (editor.cursorRect(cursor).height() - bulletWidth))
            bulletOffset = editor.contentOffset().y() + marginY
            # Prepare painter
            painter.setPen(QtGui.QColor("#777"))
            painter.setBrush(format.fore)
            painter.setRenderHint(painter.RenderHint.Antialiasing)

            # Draw breakpoints
            for linenr, (enabled, *_) in sorted(breakpoints.items()):
                blockNumber = linenr - 1
                if blockNumber < startBlockNumber:
                    continue
                # Get block
                block = editor.document().findBlockByNumber(blockNumber)
                if block.isValid():
                    y = editor.blockBoundingGeometry(block).y() + bulletOffset
                    if enabled:
                        painter.drawEllipse(
                            marginX, int(y), int(bulletWidth), int(bulletWidth)
                        )
                    else:
                        painter.drawPie(
                            marginX,
                            int(y),
                            int(bulletWidth),
                            int(bulletWidth),
                            90 * 16,
                            180 * 16,
                        )

            # Draw *the* debug marker
            if debugBlockIndicator >= 0:
                painter.setBrush(QtGui.QColor("#6F6"))
                # Get block
                block = editor.document().findBlockByNumber(debugBlockIndicator)
                if block.isValid():
                    y = editor.blockBoundingGeometry(block).y() + bulletOffset
                    y += 0.25 * bulletWidth
                    painter.drawEllipse(
                        marginX, int(y), int(bulletWidth), int(0.5 * bulletWidth)
                    )

            # Draw other debug markers
            for debugLineIndicator in editor._debugLineIndicators:
                debugBlockIndicator = debugLineIndicator - 1
                painter.setBrush(QtGui.QColor("#DDD"))
                # Get block
                block = editor.document().findBlockByNumber(debugBlockIndicator)
                if block.isValid():
                    y = editor.blockBoundingGeometry(block).y() + bulletOffset
                    y += 0.25 * bulletWidth
                    painter.drawEllipse(
                        marginX, int(y), int(bulletWidth), int(0.5 * bulletWidth)
                    )

            # Draw virtual break point
            if virtualBreakpoint > 0:
                painter.setBrush(QtGui.QColor(0, 0, 0, 0))
                # Get block
                block = editor.document().findBlockByNumber(virtualBreakpoint)
                if block.isValid():
                    y = editor.blockBoundingGeometry(block).y() + bulletOffset
                    painter.drawEllipse(
                        marginX, int(y), int(bulletWidth), int(bulletWidth)
                    )

            # Done
            painter.end()

    def __init__(self, *args, **kwds):
        self.__breakPointArea = None
        self.__leftMarginHandle = None
        super().__init__(*args, **kwds)
        # Create widget that draws the breakpoints
        self._updateBreakPointWidth()
        self.__breakPointArea = self.__BreakPointArea(self)
        self.__leftMarginHandle = self._setLeftBarMargin(
            self.__leftMarginHandle, self._getBreakPointAreaWidth()
        )
        self._breakPoints = {}  # int -> enabled, block, blockPrev, blockNext
        self._debugLineIndicator = 0
        self._debugLineIndicators = set()
        self.blockCountChanged.connect(self.__onBlockCountChanged)

    def __onBlockCountChanged(self):
        """Track breakpoints so we can update the number when text is inserted
        above.
        """
        newBreakPoints = {}
        breakPointDeleted = False

        for linenr in list(self._breakPoints):
            enabled, block, block_previous, block_next = self._breakPoints[linenr]

            # Apparently there is a bug in Qt5 and Qt6 that can cause a segmentation fault.
            # When there is a faulty block "block.previous()", calling methods such as
            # "block.previous().blockNumber()" will crash Pyzo immediately.
            # These crashes happend sometimes after Qt re-created a block and we still
            # have the outdated block reference in our self._breakPoints dictionary.
            # Such a block re-creation will happen for example during undo operations, but
            # also when a block is merged with a previous empty one (by pressing backspace
            # at the beginning of a line below an empty line).
            # See https://github.com/pyzo/pyzo/pull/949
            #
            # So, to avoid crashes, we discard breakpoints that had their block re-created.
            # To detect if the block was re-created, we check if its userData was reset
            # to None. When adding a block to "self._breakPoints", we make sure it has
            # userData set to something different than None.
            #
            # According to the Qt docs from https://doc.qt.io/qt-6/qtextblock.html:
            # "The user data object is not stored in the undo history, so it will not be
            # available after undoing the deletion of a text block."

            if block.userData() is None:
                # The block assigned to the breakpoint was re-created by Qt after we
                # added it to the self._breakPoints dict.
                # To avoid a crash we delete the breakpoint.
                del self._breakPoints[linenr]
                breakPointDeleted = True
                continue

            block_linenr = block.blockNumber() + 1
            prev_ok = block.previous().blockNumber() == block_previous.blockNumber()
            next_ok = block.next().blockNumber() == block_next.blockNumber()

            if prev_ok or next_ok:
                if block_linenr == linenr:
                    if prev_ok and next_ok:
                        pass  # All is well
                    else:
                        # Update refs
                        self._breakPoints[linenr] = (
                            enabled,
                            block,
                            block.previous(),
                            block.next(),
                        )
                else:
                    # Update linenr -- this is the only case where we "move" the breakpoint
                    newBreakPoints[block_linenr] = self._breakPoints.pop(linenr)
                    breakPointDeleted = True
            else:
                if block_linenr == linenr:
                    # Just update refs
                    self._breakPoints[linenr] = (
                        enabled,
                        block,
                        block.previous(),
                        block.next(),
                    )
                else:
                    # unexpected --> delete breakpoint
                    del self._breakPoints[linenr]
                    breakPointDeleted = True

        if newBreakPoints or breakPointDeleted:
            self._breakPoints.update(newBreakPoints)
            self.breakPointsChanged.emit(self)
            self.__breakPointArea.update()

    def breakPoints(self, triState=False):
        """A list of breakpoints for this editor."""
        if triState:
            return sorted(
                (linenr, enabled) for linenr, (enabled, *_) in self._breakPoints.items()
            )
        else:
            return sorted(
                linenr for linenr, (enabled, *_) in self._breakPoints.items() if enabled
            )

    def clearBreakPoints(self):
        """Remove all breakpoints for this editor."""
        for linenr in self.breakPoints():
            self.toggleBreakpoint(linenr, triState=False)

    def toggleBreakpoint(self, linenr=None, triState=True):
        """Turn breakpoint on / half (disabled) / off (removed) for given linenr of current line."""
        if linenr is None:
            linenr = self.textCursor().blockNumber() + 1

        if linenr in self._breakPoints:
            enabled, b, bPrev, bNext = self._breakPoints.pop(linenr)
            if triState and enabled:
                enabled = False
            else:
                b = None  # indicate to not add a new modified breakpoint
        else:
            enabled = True
            c = self.textCursor()
            c.movePosition(c.MoveOperation.Start)
            c.movePosition(c.MoveOperation.NextBlock, c.MoveMode.MoveAnchor, linenr - 1)
            b = c.block()
            bPrev = b.previous()
            bNext = b.next()

        if b is not None:
            # As described in method "__onBlockCountChanged", we want to make sure that
            # the userData of the block is set to anything but None. It is totally ok that
            # other Pyzo modules such as the highlighter will overwrite that userData as
            # long as it is not set back to None.
            if b.userData() is None:
                b.setUserData(
                    QtGui.QTextBlockUserData()
                )  # just something other than None

            self._breakPoints[linenr] = enabled, b, bPrev, bNext

        self.breakPointsChanged.emit(self)
        self.__breakPointArea.update()

    def jumpPreviousBreakpoint(self):
        currentLinenr = self.textCursor().blockNumber() + 1
        newLinenr = max(
            (linenr for linenr in self._breakPoints if linenr < currentLinenr),
            default=None,
        )
        if newLinenr is not None:
            self.gotoLine(newLinenr, keepHorizontalPos=True)

    def jumpNextBreakpoint(self):
        currentLinenr = self.textCursor().blockNumber() + 1
        newLinenr = min(
            (linenr for linenr in self._breakPoints if linenr > currentLinenr),
            default=None,
        )
        if newLinenr is not None:
            self.gotoLine(newLinenr, keepHorizontalPos=True)

    def setDebugLineIndicator(self, linenr, active=True):
        """Set the debug line indicator to the given line number.
        If None or 0, the indicator is hidden.
        """
        linenr = int(linenr or 0)
        if not linenr:
            # Remove all indicators
            if self._debugLineIndicator or self._debugLineIndicators:
                self._debugLineIndicator = 0
                self._debugLineIndicators = set()
                self.__breakPointArea.update()
        elif active:
            # Set *the* indicator
            if linenr != self._debugLineIndicator:
                self._debugLineIndicators.discard(linenr)
                self._debugLineIndicator = linenr
                self.__breakPointArea.update()
        else:
            # Add to set of indicators
            if linenr not in self._debugLineIndicators:
                self._debugLineIndicators.add(linenr)
                self.__breakPointArea.update()

    def _getBreakPointAreaWidth(self):
        if not self.__showBreakPoints:
            return 0
        else:
            return self._breakPointWidth

    def showBreakPoints(self):
        return self.__showBreakPoints

    @ce_option(True)
    def setShowBreakPoints(self, value):
        self.__showBreakPoints = bool(value)
        # Note that this method is called before the __init__ is finished,
        # so that the area is not yet created.
        self._updateBreakPointWidth()
        if self.__breakPointArea:
            if self.__showBreakPoints:
                self.__breakPointArea.show()
            else:
                self.__breakPointArea.hide()
                self.clearBreakPoints()
        if self.__leftMarginHandle is not None:
            self._setLeftBarMargin(
                self.__leftMarginHandle, self._getBreakPointAreaWidth()
            )

    def _updateBreakPointWidth(self):
        """set width of breakpoint bar (actual points are smaller)"""
        self._breakPointWidth = round(0.8 * self.cursorRect(self.textCursor()).height())

    def resizeEvent(self, event):
        super().resizeEvent(event)

        # On resize, resize the breakpointArea, too
        self._updateBreakPointWidth()
        rect = self.contentsRect()
        m = self._getMarginBeforeLeftBar(self.__leftMarginHandle)
        w = self._getBreakPointAreaWidth()
        self.__breakPointArea.setGeometry(rect.x() + m, rect.y(), w, rect.height())
        self.__leftMarginHandle = self._setLeftBarMargin(
            self.__leftMarginHandle, self._getBreakPointAreaWidth()
        )

    def paintEvent(self, event):
        super().paintEvent(event)
        # On repaint, update the complete breakPointArea
        w = self._getBreakPointAreaWidth()
        self.__breakPointArea.update(0, 0, w, self.height())


class Wrap:
    def wrap(self):
        """Enable or disable wrapping"""
        option = self.document().defaultTextOption()
        return not bool(option.wrapMode() == option.WrapMode.NoWrap)

    @ce_option(True)
    def setWrap(self, value):
        option = self.document().defaultTextOption()
        if value:
            option.setWrapMode(option.WrapMode.WrapAtWordBoundaryOrAnywhere)
        else:
            option.setWrapMode(option.WrapMode.NoWrap)
        self.document().setDefaultTextOption(option)


# todo: move this bit to base class?
# This functionality embedded in the highlighter and even has a designated
# subpackage. I feel that it should be a part of the base editor.
# Note: if we do this, remove the hasattr call in the highlighter.
class SyntaxHighlighting:
    """Notes on syntax highlighting.

    The syntax highlighting/parsing is performed using three "components".

    The base component are the token instances. Each token simply represents
    a row of characters in the text the belong to each-other and should
    be styled in the same way. There is a token class for each particular
    "thing" in the code, such as comments, strings, keywords, etc. Some
    tokens are specific to a particular language.

    There is a function that produces a set of tokens, when given a line of
    text and a state parameter. There is such a function for each language.
    These "parsers" are defined in the parsers subpackage.

    And lastly, there is the Highlighter class, that applies the parser function
    to obtain the set of tokens and using the names of these tokens applies
    styling. The styling can be defined by giving a dict that maps token names
    to style representations.

    """

    # Register all syntax style elements
    _styleElements = Manager.getStyleElementDescriptionsForAllParsers()

    def __init__(self, *args, **kwargs):
        self.__parser = None
        super().__init__(*args, **kwargs)

    def parser(self):
        """Get the parser instance currently in use to parse the code for
        syntax highlighting and source structure. Can be None.
        """
        try:
            return self.__parser
        except AttributeError:
            return None

    @ce_option(None)
    def setParser(self, parserName=""):
        """Set the current parser by giving the parser name."""
        # Set parser
        parser = Manager.getParserByName(parserName)
        if self.__parser is not parser:
            self.__parser = parser

            # We only call setStyle if the parser has changed.
            # Otherwise older shell prompts would lose some formatting.

            # Restyle, use setStyle for lazy updating
            self.setStyle()



"""compact tab widget class

See docs of the tab widget.

"""

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
import sys

ELLIPSIS = chr(8230)

# Constants for the alignments of tabs
MIN_NAME_WIDTH = 4
MAX_NAME_WIDTH = 64


## Define style sheet for the tabs

STYLESHEET = """
QTabWidget::pane { /* The tab widget frame */
    border-top: 0px solid #A09B90;
}

QTabWidget::tab-bar {
    left: 0px; /* move to the right by x px */
}


/* Style the tab using the tab sub-control. Note that
 it reads QTabBar _not_ QTabWidget */
QTabBar::tab {
    background: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                stop: 0.0 GRADIENT_UNSEL1,
                stop: 0.4 GRADIENT_UNSEL2,
                stop: 1.0 GRADIENT_UNSEL3 );
    border: 1px solid #A09B90;
    border-bottom-color: #DAD5CC; /* same as the pane color */
    border-top-left-radius: 4px;
    border-top-right-radius: 4px;
    min-width: 5ex;
    padding-bottom: PADDING_BOTTOMpx;
    padding-top: PADDING_TOPpx;
    padding-left: PADDING_LEFTpx;
    padding-right: PADDING_RIGHTpx;
    margin-right: -1px; /* "combine" borders */
}
QTabBar::tab:last {
    margin-right: 0px;
}

/* Style the selected tab, hoovered tab, and other tabs. */
QTabBar::tab:hover {
    background: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                stop: 0.0 GRADIENT_SEL1,
                stop: 0.4 GRADIENT_SEL2,
                stop: 1.0 GRADIENT_SEL3 );
}
QTabBar::tab:selected {
    background: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                stop: 0.0 GRADIENT_TOP_SELECTED,
                stop: 0.12 GRADIENT_TOP_SELECTED,
                stop: 0.120001 GRADIENT_SEL1,
                stop: 0.4 GRADIENT_SEL2,
                stop: 1.0 GRADIENT_SEL3 );
}

QTabBar::tab:selected {
    border-width: 1px;
    border-bottom-width: 0px;
    border-top-left-radius: 5px;
    border-top-right-radius: 5px;
    border-color: BORDER_COLOR_SELECTED;
}

QTabBar::tab:!selected {
    margin-top: 3px; /* make non-selected tabs look smaller */
}

"""

STYLESHEET_REPLACEMENTS = [  # name, dark, light
    ("BORDER_COLOR_SELECTED", "#ddd", "#333"),
    ("GRADIENT_TOP_SELECTED", "rgba(0,255,255,128)", "rgba(0,0,128,128)"),
    ("GRADIENT_UNSEL1", "rgba(0,0,0,128)", "rgba(220,220,220,128)"),
    ("GRADIENT_UNSEL2", "rgba(140,140,140,128)", "rgba(200,200,200,128)"),
    ("GRADIENT_UNSEL3", "rgba(160,160,160,128)", "rgba(100,100,100,128)"),
    ("GRADIENT_SEL1", "rgba(0,0,0,128)", "rgba(245,250,255,128)"),
    ("GRADIENT_SEL2", "rgba(50,50,50,128)", "rgba(210,210,210,128)"),
    ("GRADIENT_SEL3", "rgba(100,100,100,128)", "rgba(200,200,200,128)"),
]


## Define tab widget class


class TabData:
    """To keep track of real names of the tabs, but also keep supporting
    tabData.
    """

    def __init__(self, name):
        self.name = name
        self.data = None


class CompactTabBar(QtWidgets.QTabBar):
    """CompactTabBar(parent, *args, padding=(4, 4, 6, 6), preventEqualTexts=True)

    Tab bar corresponding to the CompactTabWidget.

    With the "padding" argument the padding of the tabs can be chosen.
    It should be an integer, or a 4 element tuple specifying the padding
    for top, bottom, left, right. When a tab has a button,
    the padding is the space between button and text.

    With preventEqualTexts to True, will reduce the amount of eliding if
    two tabs have (partly) the same name, so that they can always be
    distinguished.

    """

    # Add signal to be notified of double clicks on tabs
    tabDoubleClicked = QtCore.Signal(int)
    barDoubleClicked = QtCore.Signal()

    def __init__(self, *args, padding=(4, 4, 6, 6), preventEqualTexts=True):
        super().__init__(*args)

        # Put tab widget in document mode
        self.setDocumentMode(True)

        # Widget needs to draw its background (otherwise Mac has a dark bg)
        self.setDrawBase(False)
        if sys.platform == "darwin":
            self.setAutoFillBackground(True)

        # Set whether we want to prevent eliding for names that start the same.
        self._preventEqualTexts = preventEqualTexts

        # Allow moving tabs around
        self.setMovable(True)

        # Get padding
        if isinstance(padding, (int, float)):
            padding = padding, padding, padding, padding
        elif isinstance(padding, (tuple, list)):
            pass
        else:
            raise ValueError("Invalid value for padding.")

        # Set style sheet
        stylesheet = STYLESHEET
        stylesheet = stylesheet.replace("PADDING_TOP", str(padding[0]))
        stylesheet = stylesheet.replace("PADDING_BOTTOM", str(padding[1]))
        stylesheet = stylesheet.replace("PADDING_LEFT", str(padding[2]))
        stylesheet = stylesheet.replace("PADDING_RIGHT", str(padding[3]))

        for name, dark, light in STYLESHEET_REPLACEMENTS:
            stylesheet = stylesheet.replace(name, dark if pyzo.darkQt else light)

        self.setStyleSheet(stylesheet)

        # We do our own eliding
        self.setElideMode(QtCore.Qt.TextElideMode.ElideNone)

        # Make tabs wider if there's plenty space?
        self.setExpanding(False)

        # If there's not enough space, use scroll buttons
        self.setUsesScrollButtons(True)

        # When a tab is removed, select previous
        self.setSelectionBehaviorOnRemove(self.SelectionBehavior.SelectPreviousTab)

        # Init alignment parameters
        self._alignWidth = MIN_NAME_WIDTH  # Width in characters
        self._alignWidthIsReducing = False  # Whether in process of reducing

        # Create timer for aligning
        self._alignTimer = QtCore.QTimer(self)
        self._alignTimer.setInterval(10)
        self._alignTimer.setSingleShot(True)
        self._alignTimer.timeout.connect(self._alignRecursive)

    def _compactTabBarData(self, i):
        """Get the underlying tab data for tab i. Only for internal use."""

        # Get current TabData instance
        tabData = super().tabData(i)

        # If none, make it as good as we can
        if not tabData:
            name = str(super().tabText(i))
            tabData = TabData(name)
            super().setTabData(i, tabData)

        # Done
        return tabData

    ## Overload a few methods

    def mouseDoubleClickEvent(self, event):
        i = self.tabAt(event.position().toPoint())
        if i == -1:
            # There was no tab under the cursor
            self.barDoubleClicked.emit()
        else:
            # Tab was double clicked
            self.tabDoubleClicked.emit(i)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.MouseButton.MiddleButton:
            i = self.tabAt(event.position().toPoint())
            if i >= 0:
                self.parent().tabCloseRequested.emit(i)
                return
        super().mousePressEvent(event)

    def setTabData(self, i, data):
        """set the given object at the tab with index i"""
        # Get underlying python instance
        tabData = self._compactTabBarData(i)

        # Attach given data
        tabData.data = data

    def tabData(self, i):
        """get the tab data at item i. Always returns a Python object"""

        # Get underlying python instance
        tabData = self._compactTabBarData(i)

        # Return stored data
        return tabData.data

    def setTabText(self, i, text):
        """set the text for tab i."""
        tabData = self._compactTabBarData(i)
        if text != tabData.name:
            tabData.name = text
            self.alignTabs()

    def tabText(self, i):
        """get the title of the tab at index i"""
        tabData = self._compactTabBarData(i)
        return tabData.name

    ## Overload events and protected functions

    def tabInserted(self, i):
        super().tabInserted(i)

        # Is called when a tab is inserted

        # Get given name and store
        name = str(super().tabText(i))
        tabData = TabData(name)
        super().setTabData(i, tabData)

        # Update
        self.alignTabs()

    def tabRemoved(self, i):
        super().tabRemoved(i)

        # Update
        self.alignTabs()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.alignTabs()

    def showEvent(self, event):
        super().showEvent(event)
        self.alignTabs()

    ## For aligning

    def alignTabs(self):
        """Align the tab items.

        Their names are ellided if required so that
        all tabs fit on the tab bar if possible. When there is too little
        space, the QTabBar will kick in and draw scroll arrows.
        """

        # Set name widths correct (in case new names were added)
        self._setMaxWidthOfAllItems()

        # Start alignment process
        self._alignWidthIsReducing = False
        self._alignTimer.start()

    def _alignRecursive(self):
        """Recursive alignment of the items.

        The alignment process should be initiated from alignTabs().
        """

        # Only if visible
        if not self.isVisible():
            return

        # Get tab bar and number of items
        N = self.count()

        # Get right edge of last tab and left edge of corner widget
        pos1 = self.tabRect(0).topLeft()
        pos2 = self.tabRect(N - 1).topRight()
        cornerWidget = self.parent().cornerWidget()
        if cornerWidget:
            pos3 = cornerWidget.pos()
        else:
            pos3 = QtCore.QPoint(int(self.width()), 0)
        x1 = pos1.x()
        x2 = pos2.x()
        x3 = pos3.x()
        alignMargin = x3 - (x2 - x1) - 3  # Must be positive (has margin)

        # Are the tabs too wide?
        if alignMargin < 0:
            # Tabs extend beyond corner widget

            # Reduce width then
            self._alignWidth -= 1
            self._alignWidth = max(self._alignWidth, MIN_NAME_WIDTH)

            # Apply
            self._setMaxWidthOfAllItems()
            self._alignWidthIsReducing = True

            # Try again if there's still room for reduction
            if self._alignWidth > MIN_NAME_WIDTH:
                self._alignTimer.start()

        elif alignMargin > 10 and not self._alignWidthIsReducing:
            # Gap between tabs and corner widget is a bit large

            # Increase width then
            self._alignWidth += 1
            self._alignWidth = min(self._alignWidth, MAX_NAME_WIDTH)

            # Apply
            itemsElided = self._setMaxWidthOfAllItems()

            # Try again if there's still room for increment
            if itemsElided and self._alignWidth < MAX_NAME_WIDTH:
                self._alignTimer.start()
                # self._alignTimer.timeout.emit()

        else:
            pass  # margin is good

    def _getAllNames(self):
        """get a list of all (full) tab names"""
        return [self._compactTabBarData(i).name for i in range(self.count())]

    def _setMaxWidthOfAllItems(self):
        """sets the maximum width of all items now, by eliding the names

        Returns whether any items were elided.
        """

        # Get whether an item was reduced in size
        itemReduced = False

        for i in range(self.count()):
            # Get width
            w = self._alignWidth

            # Get name
            name = self._compactTabBarData(i).name

            # If it's too long, first make it shorter by stripping dir names
            if w + 1 < len(name) and "/" in name:
                name = name.split("/")[-1]

            # Check if we can reduce the name size, correct w if necessary
            if ((w + 1) < len(name)) and self._preventEqualTexts:
                # Increase w until there are no names that start the same
                allNames = self._getAllNames()
                hasSimilarNames = True
                diff = 2
                w -= 1
                while hasSimilarNames and w < len(name):
                    w += 1
                    w2 = w - (diff - 1)
                    shortName = name[:w2]
                    similarnames = [n for n in allNames if n[:w2] == shortName]
                    hasSimilarNames = len(similarnames) > 1

            # Check again, with corrected w
            if w + 1 < len(name):
                name = name[:w] + ELLIPSIS
                itemReduced = True

            # Set text now
            super().setTabText(i, name)

        # Done
        return itemReduced


class CompactTabWidget(QtWidgets.QTabWidget):
    """CompactTabWidget(parent, *args, **kwargs)

    Implements a tab widget with a tabbar that is in document mode
    and has more compact tabs that conventional tab widgets, so more
    items fit on the same space.

    Further much care is taken to ellide the names in a smart way:
      * All items are allowed the same amount of characters instead of
        that the same amount of characters is removed from all names.
      * If there are two items with the same beginning, it is made
        sure that enough characters are shown such that the names
        can be distinguished.

    The kwargs are passed to the tab bar constructor. There are a few
    keywords arguments to influence the appearance of the tabs. See the
    CompactTabBar class.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args)

        # Set tab bar
        self.setTabBar(CompactTabBar(self, **kwargs))

        # Draw tabs at the top by default
        self.setTabPosition(QtWidgets.QTabWidget.TabPosition.North)

    def setTabData(self, i, data):
        """set the given object at the tab with index i"""
        self.tabBar().setTabData(i, data)

    def tabData(self, i):
        """get the tab data at item i. Always returns a Python object"""
        return self.tabBar().tabData(i)

    def setTabText(self, i, text):
        """set the text for tab i"""
        self.tabBar().setTabText(i, text)

    def tabText(self, i):
        """get the title of the tab at index i"""
        return self.tabBar().tabText(i)


if __name__ == "__main__":
    w = CompactTabWidget()
    w.show()

    w.addTab(QtWidgets.QWidget(w), "aapenootjedopje")
    w.addTab(QtWidgets.QWidget(w), "aapenootjedropje")
    w.addTab(QtWidgets.QWidget(w), "noot en mies")
    w.addTab(QtWidgets.QWidget(w), "boom bijv een iep")
    w.addTab(QtWidgets.QWidget(w), "roosemarijnus")
    w.addTab(QtWidgets.QWidget(w), "vis")
    w.addTab(QtWidgets.QWidget(w), "vuurvuurvuur")



from ..qt import QtCore, QtGui, QtWidgets  # noqa

Qt = QtCore.Qt


class Calltip:
    _styleElements = [
        (
            "Editor.calltip",
            "The style of the calltip. ",
            "fore:#555, back:#ff9, border:1",
        )
    ]

    class __CalltipLabel(QtWidgets.QLabel):
        def __init__(self, parent):
            super().__init__(parent)

            # Start hidden
            self.hide()
            # Accept rich text
            self.setTextFormat(QtCore.Qt.TextFormat.RichText)
            # Show as tooltip
            self.setIndent(2)
            self.setWindowFlags(QtCore.Qt.WindowType.ToolTip)

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        # Create label for call tips  -- it needs a parent, otherwise it will cause a crash on Linux with Wayland
        self.__calltipLabel = self.__CalltipLabel(self)
        # Be notified of style updates
        self.styleChanged.connect(self.__afterSetStyle)

        self.cursorPositionChanged.connect(self.__onCursorPositionChanged)

        # Prevents calltips from being shown immediately after pressing
        # the escape key.
        self.__noshow = False

        self.__startcursor = None

    def __afterSetStyle(self):
        format = self.getStyleElementFormat("editor.calltip")
        ss = "QLabel {{ color:{}; background:{}; border:{}px solid {}; }}".format(
            format["fore"],
            format["back"],
            int(format["border"]),
            format["fore"],
        )
        self.__calltipLabel.setStyleSheet(ss)

    def setCalltipFinishedCallback(self, cb):
        self.__finishedCallback = cb

    def calltipShow(self, offset=0, richText="", highlightFunctionName=False):
        """Shows the given calltip.

        Parameters
        ----------
        offset : int
            The character offset to show the tooltip.
        richText : str
            The text to show (may contain basic html for markup).
        highlightFunctionName : bool
            If True the text before the first opening brace is made bold.
            default False.
        """

        # Do not show the calltip if it was deliberately hidden by the
        # user.
        if self.__noshow:
            return

        # Process calltip text?
        if highlightFunctionName:
            i = richText.find("(")
            if i > 0:
                richText = "<b>{}</b>{}".format(richText[:i], richText[i:])

        # Get a cursor to establish the position to show the calltip
        startcursor = self.textCursor()
        startcursor.movePosition(startcursor.MoveOperation.Left, n=offset)
        self.__startcursor = startcursor

        # Get position in pixel coordinates
        rect = self.cursorRect(startcursor)
        pos = rect.topLeft()
        pos.setY(pos.y() - rect.height() - 1)  # Move one above line
        pos.setX(pos.x() - 3)  # Correct for border and indent
        pos = self.viewport().mapToGlobal(pos)

        label = self.__calltipLabel
        textChanged = richText != label.text()

        # Set text and update font
        label.setText(richText)
        label.setFont(self.font())

        # Use a qt tooltip to show the calltip
        if richText:
            if textChanged and label.isVisible():
                # When the tooltip label is still shown and we update the text, the
                # rectangle of the label is not updated. This leads to truncated text
                # or to empty space on the right of the label.
                # As a workaround, we hide and then show the label again.
                label.hide()

            label.move(pos)
            label.show()
        else:
            label.hide()

    def calltipCancel(self):
        """Hides the calltip."""
        self.__calltipLabel.hide()
        self.__startcursor = None

        if self.__finishedCallback is not None:
            self.__finishedCallback()

    def calltipActive(self):
        """Get whether the calltip is currently active."""
        return self.__calltipLabel.isVisible()

    def focusOutEvent(self, event):
        super().focusOutEvent(event)
        self.__calltipLabel.hide()

    def keyPressEvent(self, event):
        # If the user presses Escape and the calltip is active, hide it
        if (
            event.key() == Qt.Key.Key_Escape
            and event.modifiers() == Qt.KeyboardModifier.NoModifier
            and self.calltipActive()
        ):
            self.calltipCancel()
            self.__noshow = True
            return

        if event.text() == "(":
            self.__noshow = False
        elif event.text() == ")":
            self.calltipCancel()

        # Proceed processing the keystrike
        super().keyPressEvent(event)

    def __onCursorPositionChanged(self, *args):
        if self.calltipActive():
            if self.__startcursor is not None:
                if self.__startcursor.blockNumber() != self.textCursor().blockNumber():
                    self.calltipCancel()



"""Simple script to make a performance plot of the speed for sending
different package sizes.
"""

# Go up one directory and then import the codeeditor package
import os
import sys

os.chdir("../..")
sys.path.insert(0, ".")

# Import yoton from there
import yoton

# Normal imports
import time

## Run experiment with different message sizes

# Setup host
ct1 = yoton.Context()
ct1.bind("localhost:test")
c1 = yoton.PubChannel(ct1, "speedTest")

# Setup client
ct2 = yoton.SimpleSocket()
ct2.connect("localhost:test")
c2 = yoton.SubChannel(ct2, "speedTest")

# Init
minSize, maxSize = 2, 100 * 2**20
BPS = []
TPM = []
N = []
SIZE = []


# Loop
size = minSize
while size < maxSize:
    # Calculate amount
    n = int(200 * 2**20 / size)
    n = min(yoton.core.BUF_MAX_LEN, n)

    t0 = time.time()

    # Send messages
    message = "T" * int(size)
    for i in range(n):
        c1.send(message)
    ct1.flush(20.0)

    t1 = time.time()

    # In the mean while two threads are working their asses off to pop
    # the packages from one queue, send them over a socket, and push
    # them on another queue.

    # Receive messages
    for i in range(n):
        c2.recv()

    t2 = time.time()

    # Calculate speed
    etime = t2 - t0
    bps = n * size / etime  # bytes per second
    tpm = etime / n

    # Make strings
    bps_ = "%1.2f B/s" % bps
    size_ = "%i B" % size
    #
    D = {2**10: "KB", 2**20: "MB", 2**30: "GB"}
    for factor in D:
        if bps >= factor:
            bps_ = "%1.2f %s/s" % (bps / factor, D[factor])
        if size >= factor:
            size_ = "%1.2f %s" % (size / factor, D[factor])

    # Show result
    print("Sent %i messages of %s in %1.2f s: %s" % (n, size_, etime, bps_))

    # Store stuff
    N.append(n)
    SIZE.append(size)
    BPS.append(bps)
    TPM.append(tpm)

    # Prepare for next round
    size *= 1.9


## Visualize

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter

plt.ion()

fig, (ax1, ax2) = plt.subplots(2, 1, num=10, clear=True, sharex=True)


def format_si(value, unit):
    prefixes = ["micro", "m", "", "k", "M", "G"]
    i3 = int(np.log10(value) // 3)
    i = i3 + prefixes.index("")
    i = max(0, min(i, len(prefixes) - 1))
    multiplier = 10 ** (3 * i3)
    return "{:.0f} {}{}".format(value / multiplier, prefixes[i], unit)


formatter = FuncFormatter(lambda val, tick_pos: format_si(val, ""))

ax1.loglog(SIZE, BPS, ".")
ax1.set_ylabel("speed [bytes/s]")
ax2.loglog(SIZE, TPM, ".")
ax2.set_ylabel("time per message [s]")
ax2.set_xlabel("message size [bytes]")

ax2.xaxis.set_major_formatter(formatter)
ax2.tick_params("x", rotation=45)

for ax in [ax1, ax2]:
    ax.grid(True)
    ax.yaxis.set_major_formatter(formatter)

fig.tight_layout()

# fig.savefig("yoton_performance.jpg")



"""Module splash

Defines splash window shown during startup.

"""

import os

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo import translate

STYLESHEET = """
QWidget {
    background-color: #268bd2;
}
QFrame {
    background-image: url("{IMAGE_URL}");
    background-repeat: no-repeat;
    background-position: center;
}
QLabel {
    color: #222;
    background: #46abf2;
    border-radius:20px;
}
"""

splash_text = """
<p>{text_title}</p>
<p>{text_version} {version}</p>
<p>{text_os} <a href='https://pyzo.org/'>https://pyzo.org</a></p>
"""


class LogoWidget(QtWidgets.QFrame):
    def __init__(self, parent):
        super().__init__(parent)
        self.setMinimumSize(256, 256)
        self.setMaximumSize(256, 256)


class LabelWidget(QtWidgets.QWidget):
    def __init__(self, parent, distro=None):
        super().__init__(parent)
        self.setMinimumSize(360, 256)  # Ensure title fits nicely

        # Create label widget and costumize
        self._label = QtWidgets.QLabel(self)
        self._label.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self._label.setOpenExternalLinks(True)
        self._label.setWordWrap(True)
        self._label.setMargin(20)

        # Set font size (absolute value)
        font = self._label.font()
        font.setPointSize(11)  # (font.pointSize()+1)
        self._label.setFont(font)

        # Build
        text_title = translate(
            "splash", "This is <b>Pyzo</b><br />the Python IDE for scientific computing"
        )
        text_version = translate("splash", "Version")
        text_os = translate(
            "splash", "Pyzo is open source software and freely available for everyone."
        )
        text = splash_text.format(
            version=pyzo.__version__,
            text_title=text_title,
            text_version=text_version,
            text_os=text_os,
        )

        # Set text
        self._label.setText(text)

        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)
        layout.addStretch(1)
        layout.addWidget(self._label, 0)
        layout.addStretch(1)


class SplashWidget(QtWidgets.QWidget):
    """A splash widget."""

    def __init__(self, parent, **kwargs):
        super().__init__(parent)

        self._left = LogoWidget(self)
        self._right = LabelWidget(self, **kwargs)

        # Layout
        layout = QtWidgets.QHBoxLayout(self)
        self.setLayout(layout)
        # layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(25)
        layout.addStretch(1)
        layout.addWidget(self._left, 0)
        layout.addWidget(self._right, 0)
        layout.addStretch(1)

        # Change background of main window to create a splash-screen-effect
        iconImage = "pyzologo256.png"
        iconImage = os.path.join(pyzo.pyzoDir, "resources", "appicons", iconImage)
        iconImage = iconImage.replace(os.path.sep, "/")  # Fix for Windows
        self.setStyleSheet(STYLESHEET.replace("{IMAGE_URL}", iconImage))


if __name__ == "__main__":
    w = SplashWidget(None, distro="some arbitrary distro")
    w.resize(800, 600)
    w.show()



# -*- coding: utf-8 -*-

# This example defines a message type and sends such a message over pub/sub.
# This example can be run in one go or in two parts running in
# different processes.


## ========== One end

import yoton

verbosity = 0


# Create custom message type. (should be defined at both ends)
class NumberMessageType(yoton.MessageType):
    def message_from_bytes(self, bb):
        return float(bb.decode("utf-8"))

    def message_to_bytes(self, number):
        return str(number).encode("utf-8")

    def message_type_name(self):
        return "num"


# Create context, a channel, and connect
ct1 = yoton.Context(verbose=verbosity)
pub = yoton.PubChannel(ct1, "numbers", NumberMessageType)
ct1.bind("publichost:test")

# Send a message
pub.send(42.9)


## ========== Other end

import yoton

verbosity = 0


# Create custom message type. (should be defined at both ends)
class NumberMessageType(yoton.MessageType):
    def message_from_bytes(self, bb):
        return float(bb)

    def message_to_bytes(self, number):
        return str(number).encode("utf-8")

    def message_type_name(self):
        return "num"


# Create a context, a channel, and connect
ct2 = yoton.Context(verbose=verbosity)
sub = yoton.SubChannel(ct2, "numbers", NumberMessageType)
ct2.connect("publichost:test")


# Duplicate a string and a number
print(sub.recv())



"""CodeEditor

A full featured code editor component based on QPlainTextEdit.

"""

# ruff: noqa: F401

from .manager import Manager
from .base import CodeEditorBase

from .extensions.appearance import (
    HighlightMatchingBracket,
    HighlightMatchingOccurrences,
    HighlightCurrentLine,
    FullUnderlines,
    IndentationGuides,
    CodeFolding,
    LongLineIndicator,
    ShowWhitespace,
    ShowLineEndings,
    Wrap,
    LineNumbers,
    SyntaxHighlighting,
    BreakPoints,
)
from .extensions.behaviour import (
    Indentation,
    HomeKey,
    EndKey,
    NumpadPeriodKey,
    AutoIndent,
    PythonAutoIndent,
    SmartCopyAndPaste,
    DeflectUpDown,
    MoveLinesUpDown,
    ScrollWithUpDownKeys,
    AutoCloseQuotesAndBrackets,
)
from .extensions.autocompletion import AutoCompletion
from .extensions.calltip import Calltip


# Order of superclasses: first the extensions, then CodeEditorBase
# The first superclass is the first extension that gets to handle each key and
# the first to receive paint events.
class CodeEditor(
    HighlightCurrentLine,
    HighlightMatchingOccurrences,
    HighlightMatchingBracket,
    FullUnderlines,
    IndentationGuides,
    CodeFolding,
    LongLineIndicator,
    ShowWhitespace,
    ShowLineEndings,
    Wrap,
    BreakPoints,
    LineNumbers,
    AutoCompletion,  # Escape: first remove autocompletion,
    Calltip,  # then calltip
    Indentation,
    DeflectUpDown,
    MoveLinesUpDown,
    ScrollWithUpDownKeys,
    HomeKey,
    EndKey,
    # NumpadPeriodKey,  -> disabled, see issue #720
    AutoIndent,
    PythonAutoIndent,
    AutoCloseQuotesAndBrackets,
    SyntaxHighlighting,
    SmartCopyAndPaste,  # overrides cut(), copy(), paste()
    CodeEditorBase,  # CodeEditorBase must be the last one in the list
):
    """
    CodeEditor with all the extensions
    """

    pass



import os
import datetime

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa


class CommandHistory(QtCore.QObject):
    """Keep track of a (global) history of commands.

    This kinda assumes Python commands, but should be easy enough to generalize.
    """

    max_commands = 2000

    command_added = QtCore.Signal(str)
    command_removed = QtCore.Signal(int)
    commands_reset = QtCore.Signal()

    def __init__(self, fname):
        super().__init__()
        self._commands = []
        self._fname = fname
        self._last_date = datetime.date(2000, 1, 1)
        self._load()

    def _load(self):
        """Load commands from file."""
        if not self._fname:
            return

        assert not self._commands

        try:
            filename = os.path.join(pyzo.appDataDir, self._fname)
            if not os.path.isfile(filename):
                with open(filename, "wb"):
                    pass

            # Load lines and add to commands
            with open(filename, "rt", encoding="utf-8") as fd:
                lines = fd.read().splitlines()
            self._commands.extend(
                [line.rstrip() for line in lines[-self.max_commands :]]
            )

            # Resolve last date
            for c in reversed(lines):
                if c.startswith("# ==== "):
                    try:
                        c = c.split("====")[1].strip()
                        self._last_date = datetime.datetime.strptime(
                            c, "%Y-%m-%d"
                        ).date()
                        break
                    except Exception:
                        pass

        except Exception as e:
            print("An error occurred while loading the history: " + str(e))

    def save(self):
        """Save the commands to disk."""
        if not self._fname:
            return
        filename = os.path.join(pyzo.appDataDir, self._fname)
        try:
            with open(filename, "wt", encoding="utf-8") as f:
                f.write("\n".join(self._commands))
        except Exception:
            print("Could not save command history")

    def get_commands(self):
        """Get a list of all commands (latest last)."""
        return self._commands.copy()

    def append(self, command):
        """Add a command to the list."""
        command = command.rstrip()
        if not command:
            return

        # Add date?
        today = datetime.date.today()
        if today > self._last_date:
            self._last_date = today
            self._commands.append("# ==== " + today.strftime("%Y-%m-%d"))
            self.command_added.emit(self._commands[-1])

        # Clear it
        try:
            index = self._commands.index(command)
        except ValueError:
            pass
        else:
            self._commands.pop(index)
            self.command_removed.emit(index)

        # Append
        self._commands.append(command)
        self.command_added.emit(self._commands[-1])

        # Reset?
        if len(self._commands) > self.max_commands:
            self._commands[: self.max_commands // 2] = []
            self.commands_reset.emit()

    def pop(self, index):
        """Remove a command by index."""
        self._commands.pop(index)
        self.command_removed.emit(index)

    def find_starting_with(self, firstpart, n=1):
        """Find the nth (1-based) command that starts with firstpart, or None."""
        count = 0
        for c in reversed(self._commands):
            if c.startswith(firstpart):
                count += 1
                if count >= n:
                    return c
        return None

    def find_all(self, needle):
        """Find all commands that contain the given text. In order of being used."""
        return [cmd for cmd in reversed(self._commands) if needle in cmd]



import sys
import yoton
import inspect  # noqa - used in eval()


# we keep this file in ascii encoding to stay compatible with Python 2.7 kernels
THREE_DOTS_CHAR = b"\xe2\x80\xa6".decode("utf-8")

_getReprPrefixPostfix = {
    "list": ("[", "]"),
    "tuple": ("(", ")"),
    "dict": ("{", "}"),
    "frozendict": ("frozendict(", ")"),
    "set": ("{", "}"),  # only for non-empty sets
    "frozenset": ("frozenset(", ")"),
}


def getRepr(obj, maxChars, _path=(), _sList=None):
    """like repr(obj)[:maxChars], but more efficient for large objects"""

    if _sList is None:
        # the first list element is the total string length of all elements
        _sList = [0]
    tn = obj.__class__.__name__
    if tn in _getReprPrefixPostfix:
        prefix, postfix = _getReprPrefixPostfix[tn]
        if tn == "set" and not obj:
            prefix, postfix = "set(", ")"

        _sList.append(prefix)
        _sList[0] += len(prefix)
        _pathSub = _path + (obj,)
        if obj in _path:
            _sList.append("...")
            _sList[0] += 3
        else:
            for i, subObj in enumerate(obj):
                if i > 0:
                    _sList.append(", ")
                    _sList[0] += 2
                if _sList[0] >= maxChars:
                    break

                getRepr(subObj, maxChars, _pathSub, _sList)
                if tn in ("dict", "frozendict"):
                    _sList.append(": ")
                    _sList[0] += 2
                    getRepr(obj[subObj], maxChars, _pathSub, _sList)
        _sList.append(postfix)
        _sList[0] += len(postfix)
    elif hasattr(obj, "__len__"):
        # e.g.: str, bytes, bytearray
        s = repr(obj[:maxChars])
        _sList.append(s)
        _sList[0] += len(s)
    else:
        s = repr(obj)
        _sList.append(s)
        _sList[0] += len(s)

    if not _path:
        return "".join(_sList[1:])[:maxChars]


# Functions/objects like "dir" could be redefined in the namespace of the user's code.
# To use our known "dir", we assign it to a rather unique variable name that is only
# visible in a global namespace used for introspection, e.g.:
#   dir --> __pyzo_internalNS_dir__

INTERNAL_OBJECT_NAMES = [
    "inspect",
    "hasattr",
    "dir",
    "str",
    "getRepr",
]

internalNS = {}  # namespace for eval/exec (look-up table for unique_name to object)
inames = {}  # look-up table for name to unique_name
for name in INTERNAL_OBJECT_NAMES:
    uniqueName = "__pyzo_internalNS_{}__".format(name)
    internalNS[uniqueName] = eval(name)
    inames[name] = uniqueName


class PyzoIntrospector(yoton.RepChannel):
    """This is a RepChannel object to respond to requests from the IDE."""

    def _getNameSpace(self, name=""):
        """Get the namespace to apply introspection in.

        If name is given, will find that name. For example sys.stdin.
        """

        # Get namespace
        NS1 = sys._pyzoInterpreter.locals
        NS2 = sys._pyzoInterpreter.globals
        if not NS2:
            NS = NS1
        else:
            NS = NS2.copy()
            NS.update(NS1)

        # Look up a name?
        if not name:
            return NS
        else:
            try:
                # Get object
                ob = eval(name, None, NS)

                # Get namespace for this object
                if isinstance(ob, dict) or hasattr(
                    ob, "keys"
                ):  # os.environ is no dict but has keys
                    NS = {"[" + repr(el) + "]": ob[el] for el in ob.keys()}
                elif isinstance(ob, (list, tuple)):
                    NS = {}
                    for i, el in enumerate(ob):
                        if i >= 1000:
                            NS["[{}]".format(len(ob) - 1)] = ob[-1]  # add last object
                            NS["# ... too many entries"] = None
                            break
                        NS["[{}]".format(i)] = el
                else:
                    keys = dir(ob)
                    NS = {}
                    for key in keys:
                        try:
                            NS[key] = getattr(ob, key)
                        except Exception:
                            NS[key] = "<unknown>"

                # Done
                return NS

            except Exception:
                return {}

    def _getSignature(self, objectName):
        """Get the signature of builtin, function or method.

        Returns a tuple (signature_string, kind), where kind is a string
        of one of the above. When none of the above, both elements in
        the tuple are an empty string.
        """

        # if a class, get init
        # not if an instance! -> try __call__ instead
        # what about self?

        # Get valid object names
        parts = objectName.rsplit(".")
        objectNames = [".".join(parts[-i:]) for i in range(1, len(parts) + 1)]

        # find out what kind of function, or if a function at all!
        originalObjectName = objectName
        NS = self._getNameSpace()

        if not eval(
            "{}({}, '__call__')".format(inames["hasattr"], objectName), internalNS, NS
        ):
            # Return early if the object not is callable.
            # With this we avoid unnecessary delays and CPU and memory usage.
            #
            # For example, when executing
            #   inspect.signature(obj)
            # where obj is a very long list, an exception
            #   f"TypeError: {repr(obj)} is not a callable object'
            # will be raised, and executing "repr(obj)" of a very long list could take a
            # few seconds and consume a lot of memory.
            return "", ""

        fun1 = eval(inames["inspect"] + ".isbuiltin(%s)" % (objectName), internalNS, NS)
        fun2 = eval(
            inames["inspect"] + ".isfunction(%s)" % (objectName), internalNS, NS
        )
        fun3 = eval(inames["inspect"] + ".ismethod(%s)" % (objectName), internalNS, NS)
        fun4 = False
        fun5 = False
        if not (fun1 or fun2 or fun3):
            # Maybe it's a class with an init?
            if eval(
                inames["hasattr"] + "(%s, '__init__')" % (objectName), internalNS, NS
            ):
                objectName += ".__init__"
                fun4 = eval(
                    inames["inspect"] + ".ismethod(%s)" % (objectName), internalNS, NS
                )
            else:
                objectName += ".__call__"
                fun5 = eval(
                    inames["inspect"] + ".ismethod(%s)" % (objectName), internalNS, NS
                )

        sigs = ""
        if True:
            # the first line in the docstring is usually the signature
            tmp = eval("%s.__doc__" % (objectNames[-1]), {}, NS)
            sigs = ""
            if tmp:
                # we need the lstrip for docstrings like the one of np.where:
                #   np.where.__doc__[:34] == '\n    where(condition, [x, y], /)\n\n'
                tmp = tmp.lstrip()

                # we will try joining several lines because doc strings of objects
                # like np.array have a new line inside the signature
                for line in tmp.splitlines()[:3]:
                    line = line.strip()
                    sigs += line
                    if not line.endswith(","):
                        break
                    sigs += " "
            # Test if doc has signature
            hasSig = False
            for name in objectNames:  # list.append -> L.append(object) -- blabla
                name += "("
                if name in sigs:
                    hasSig = True
            if not hasSig and sys.version_info[0] >= 3:
                # Also get the signature for objects where inspect.signature does not
                # work and where the function/method has a differnt name, e.g.:
                # import sys; myfunc = sys.getsizeof
                # myfunc.__doc__ --> has a docstring starting with "getsizeof("
                indOpenParen = sigs.find("(")
                if indOpenParen > 0 and sigs[:indOpenParen].isidentifier():
                    hasSig = True
                    sigs = objectName + sigs[indOpenParen:]
            # If not a valid signature, do not bother ...
            if (not hasSig) or (sigs.count("(") != sigs.count(")")):
                sigs = ""
            if sigs == "":
                # try inspect.signature to also get a signature for objects like
                #   np.sum, pd.DataFrame, ...
                try:
                    sigs = eval(
                        inames["str"]
                        + "("
                        + inames["inspect"]
                        + ".signature(%s))" % (originalObjectName),
                        internalNS,
                        NS,
                    )
                    sigs = originalObjectName + sigs
                except Exception:
                    pass

        if fun1 or fun2 or fun3 or fun4 or fun5:
            if fun1:
                kind = "builtin"
            elif fun2:
                kind = "function"
            elif fun3:
                kind = "method"
            elif fun4:
                kind = "class"
            elif fun5:
                kind = "callable"

            if not sigs:
                # Use introspection

                funname = objectName.split(".")[-1]

                try:
                    tmp = eval(
                        inames["inspect"] + ".signature(%s)" % (objectName),
                        internalNS,
                        NS,
                    )  # py3.3
                    sigs = funname + str(tmp)
                except Exception:
                    try:
                        # for Python < v3.3
                        tmp = eval(
                            inames["inspect"] + ".getargspec(%s)" % (objectName),
                            internalNS,
                            NS,
                        )
                    except Exception:  # the above fails for builtins
                        tmp = None
                        kind = ""

                    if tmp is not None:
                        args, varargs, varkw, defaults = tmp[:4]
                        # prepare defaults
                        if defaults is None:
                            defaults = ()
                        defaults = list(defaults)
                        defaults.reverse()
                        # make list (back to forth)
                        args2 = []
                        for i in range(len(args) - fun4):
                            arg = args.pop()
                            if i < len(defaults):
                                args2.insert(0, "%s=%s" % (arg, defaults[i]))
                            else:
                                args2.insert(0, arg)
                        # append varargs and kwargs
                        if varargs:
                            args2.append("*" + varargs)
                        if varkw:
                            args2.append("**" + varkw)
                        # append the lot to our string
                        sigs = "%s(%s)" % (funname, ", ".join(args2))

        elif sigs:
            kind = "function"
        else:
            sigs = ""
            kind = ""

        return sigs, kind

    def dirWithIntermediateResult(self, objectName):
        cacheVarName = "__pyzo__autocomp"
        try:
            command = cacheVarName + " = " + str(objectName)
            exec(command, sys._pyzoInterpreter.globals, sys._pyzoInterpreter.locals)
        except Exception:
            return []
        else:
            return self.dir(cacheVarName)

    # todo: variant that also says whether it's a property/function/class/other
    def dir(self, objectName):
        """Get list of attributes for the given name."""
        # sys.__stdout__.write('handling '+objectName+'\n')
        # sys.__stdout__.flush()

        # Get namespace
        NS = self._getNameSpace()

        # Init names
        names = set()

        # Obtain all attributes of the class
        try:
            command = inames["dir"] + "(%s.__class__)" % (objectName)
            d = eval(command, internalNS, NS)
        except Exception:
            pass
        else:
            names.update(d)

        # Obtain instance attributes
        try:
            command = "%s.__dict__.keys()" % (objectName)
            d = eval(command, {}, NS)
        except Exception:
            pass
        else:
            names.update(d)

        # That should be enough, but in case __dir__ is overloaded,
        # query that as well
        try:
            command = inames["dir"] + "(%s)" % (objectName)
            d = eval(command, internalNS, NS)
        except Exception:
            pass
        else:
            names.update(d)

        # Respond
        return list(names)

    def dir2WithIntermediateResult(self, objectName):
        cacheVarName = "__pyzo__autocomp"
        try:
            command = cacheVarName + " = " + str(objectName)
            exec(command, sys._pyzoInterpreter.globals, sys._pyzoInterpreter.locals)
        except Exception:
            return []
        else:
            return self.dir2(cacheVarName)

    def dir2(self, objectName):
        """Get variable names in currently active namespace plus extra information.

        Returns a list of tuple of strings: name, type, kind, repr.
        """
        try:
            name = ""
            names = []

            def storeInfo(name, val):
                # Determine type
                typeName = type(val).__name__
                # Determine kind
                kind = typeName
                if typeName != "type":
                    if (
                        hasattr(val, "__array__")
                        and hasattr(val, "dtype")
                        and hasattr(val, "shape")
                        and not hasattr(
                            val,
                            "if_this_is_an_attribute_then_there_are_likely_inf_attributes",
                        )
                    ):
                        kind = "array"
                    elif isinstance(val, list):
                        kind = "list"
                    elif isinstance(val, tuple):
                        kind = "tuple"
                # Determine representation
                if kind == "array":
                    tmp = "x".join([str(s) for s in val.shape])
                    if tmp:
                        values_repr = ""
                        if hasattr(val, "flat"):
                            for el in val.flat:
                                # using str instead of repr to have
                                # "<array 3 int64: 1, 2, 3>"
                                # instead of
                                # "<array 3 int64: np.int64(1), np.int64(2), np.int64(3)>"
                                values_repr += ", " + str(el)
                                if len(values_repr) > 70:
                                    values_repr = values_repr[:69] + THREE_DOTS_CHAR
                                    break
                        repres = "<array %s %s: %s>" % (
                            tmp,
                            val.dtype.name,
                            values_repr[2:],  # remove the leading ", "
                        )
                    elif val.size:
                        # val can be a non-numeric or structured type as well, e.g.:
                        #     val = np.array(['abc'], dtype=np.dtype('U3'))[0]
                        repres = "<array scalar %s (%s)>" % (val.dtype.name, str(val))
                    else:
                        repres = "<array empty %s>" % (val.dtype.name)
                elif kind in (
                    "list",
                    "tuple",
                    "dict",
                    "frozendict",
                    "set",
                    "frozenset",
                ):
                    maxChars = 70
                    values_repr = getRepr(val, maxChars + 1)
                    if len(values_repr) > maxChars:
                        values_repr = values_repr[: maxChars - 1] + THREE_DOTS_CHAR

                    if kind in ("frozendict", "frozenset"):
                        values_repr = values_repr[len(kind + "({") : -1]
                    else:
                        values_repr = values_repr[1:-1]

                    if kind in ("dict", "frozendict"):
                        elem = "item"
                    else:
                        elem = "element"

                    repres = "<{}-{} {}: {}>".format(len(val), elem, kind, values_repr)
                else:
                    maxChars = 80
                    repres = getRepr(val, maxChars + 1)
                    if len(repres) > maxChars:
                        repres = repres[: maxChars - 1] + THREE_DOTS_CHAR
                # Store
                tmp = (name, typeName, kind, repres)
                names.append(tmp)

            # Get locals
            NS = self._getNameSpace(objectName)
            for name in NS.keys():  # name can be a key in a dict, i.e. not str
                if hasattr(name, "startswith") and name.startswith("__"):
                    continue
                try:
                    storeInfo(str(name), NS[name])
                except Exception:
                    pass

            return names

        except Exception:
            return []

    def signatureWithIntermediateResult(self, objectName):
        cacheVarName = "__pyzo__calltip"
        try:
            command = cacheVarName + " = " + str(objectName)
            exec(command, sys._pyzoInterpreter.globals, sys._pyzoInterpreter.locals)
        except Exception:
            return None
        else:
            return self.signature(cacheVarName)

    def signature(self, objectName):
        """Get signature."""
        try:
            text, kind = self._getSignature(objectName)
            return text
        except Exception:
            return None

    def doc(self, objectName):
        """Get documentation for an object."""

        # Get namespace
        NS = self._getNameSpace()

        try:
            if objectName.endswith("[# ... too many entries]"):
                raise IndexError()
            # collect docstring
            h_text = ""
            # Try using the class (for properties)
            try:
                className = eval("%s.__class__.__name__" % (objectName), {}, NS)
                if "." in objectName:
                    tmp = objectName.rsplit(".", 1)
                    tmp[1] += "."
                else:
                    tmp = [objectName, ""]
                if className not in [
                    "type",
                    "module",
                    "builtin_function_or_method",
                    "function",
                ]:
                    cmd = "%s.__class__.%s__doc__"
                    h_text = eval(cmd % (tmp[0], tmp[1]), {}, NS)
            except Exception:
                pass

            # Normal doc
            if not h_text:
                h_text = eval("%s.__doc__" % (objectName), {}, NS)

            # collect more data
            maxChars = 200
            h_repr = eval(
                "{}({}, {})".format(inames["getRepr"], objectName, maxChars + 1),
                internalNS,
                NS,
            )
            if len(h_repr) > maxChars:
                h_repr = h_repr[: maxChars - 1] + THREE_DOTS_CHAR

            try:
                h_class = eval("%s.__class__.__name__" % (objectName), {}, NS)
            except Exception:
                h_class = "unknown"

            # docstring can be None, but should be empty then
            if not h_text:
                h_text = ""

            # get and correct signature
            h_fun, kind = self._getSignature(objectName)

            if not h_fun:
                h_fun = ""  # signature not available

            # replace newlines so we can separates the different parts
            h_repr = h_repr.replace("\n", "\r")

            # build final text
            text = "\n".join([objectName, h_class, h_fun, h_repr, h_text])

        except Exception:
            type, value, tb = sys.exc_info()
            del tb
            text = "\n".join(
                [objectName, "", "", "", "No help available. ", str(value)]
            )

        # Done
        return text

    def evalMultiple(self, sequenceOfExpressions):
        """evaluate multiple expressions

        The result will be converted to a string and truncated if too long.
        In case of an error, the result will be the exception as a string.

        return value:
            [(request1, success1, result1), (request2, success2, result2), ...]
        """
        NS = self._getNameSpace()
        maxChars = 150
        returnValue = []
        for expr in sequenceOfExpressions:
            exprOrig = expr
            if expr.startswith("repr(") and expr.endswith(")"):
                expr = "{}({}, {})".format(inames["getRepr"], expr[5:-1], maxChars + 1)
            try:
                result = str(eval(expr, internalNS, NS))
                if len(result) > maxChars:
                    result = result[: maxChars - 1] + THREE_DOTS_CHAR
                success = True
            except Exception as e:
                result = str(e)
                success = False

            returnValue.append((exprOrig, success, result))

        return returnValue

    def eval(self, command):
        """Evaluate a command and return result."""

        # Get namespace
        NS = self._getNameSpace()

        try:
            # here globals is None, so we can look into sys, time, etc...
            return eval(command, None, NS)
        except Exception:
            return "Error evaluating: " + command


##
if False:
    ## execute this cell to test signature extraction
    tt = """
    bytes
    int
    iter
    min
    import numpy as np; np.linspace; np.sum; np.array; np.linalg.norm
    import pandas as pd; pd.DataFrame
    import requests; requests.Request
    """
    for line in tt.strip().splitlines():
        ll = line.split("; ")
        if len(ll) == 1:
            ll.insert(0, None)
        if ll[0] is not None:
            exec(ll[0].strip())
        for s in ll[1:]:
            sig = __pyzo__.introspector.signature(s)  # noqa: F821
            if sig == "":
                print("!!! could not extract the signature of", s, "!!!")
            else:
                print(sig)
            print()



"""
This is an example for using externally started shells with Pyzo.
Some third party software start their own Python interpreter process.
To use that Python process in a shell in Pyzo, just have Pyzo running
and then execute this code in the external Python interpreter.

For example, start a normal Python interpreter in a terminal resp. command window so that
you have an interactive Python session.
In this Python session, execute the following code lines:

fp = '/path/to/start_external_shell.py'
with open(fp, 'rt') as fd: pyzo_kernel_code = fd.read()
exec(pyzo_kernel_code)

This will create a shell in Pyzo's "Shells" panel. You can now use it like a shell that
was started inside Pyzo. To give back control to the external Python interpreter, just
close the shell inside Pyzo (via "Shell -> Close" in Pyzo's menu).
After closing the shell, you can start the Pyzo kernel again by executing
exec(pyzo_kernel_code)
again. All variables are still there because it is the same Python session.

A more real-world example, where external shells can be useful:
Run the software "Blender" -- https://en.wikipedia.org/wiki/Blender_(software)
and enter the code above in Blender's Python console to run the Pyzo kernel.
Now you can use Pyzo to perform commands as listed below, though the Blender GUI will
freeze till the shell is closed again. But you can re-enter the kernel anytime.
bpy.ops.mesh.primitive_cube_add(size=3, location=(0, 0, 0))
bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
# https://docs.blender.org/api/current/info_gotcha.html#can-i-redraw-during-script-execution

The global and local variables of the new shell are the ones of the outermost frame.
But they can be switched to other frames, for example to frames between the outmost frame
and the Pyzo kernel's first frame. To do this, execute for example the line
    import sys; sys._pyzoInterpreter.switchframe(1)
which switches globals and locals to the second outermost frame (index 1).
"""


def _run_pyzo_kernel():

    import os
    import sys
    import ast
    import socket
    import textwrap

    def send_data_to_pyzo_commandserver(shell_config):
        # TCP client
        remote_port = 63859  # port_hash('pyzoserver') --> 63859
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(1.0)
            try:
                sock.connect(('127.0.0.1', remote_port))
                sock.sendall('startexternalshell {!r}\r\n'.format(shell_config).encode('utf-8'))
                data = sock.recv(1024)
                answer = data.decode('utf-8')
                return answer
            except socket.timeout as e:
                print('could not connect to the pyzo commandserver')
                raise e

    def get_pyzo_shell_interface(shell_config):
        # TCP server
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.bind(('127.0.0.1', 0))
            my_server_port = sock.getsockname()[1]
            sock.settimeout(5.0)
            sock.listen(1)

            shell_config['externalshell_callbackport'] = my_server_port
            answer = send_data_to_pyzo_commandserver(shell_config)
            if answer != 'ok\r\n':
                raise ValueError('invalid response from commandserver')

            sock_conn, addr = sock.accept()
            sock_conn.settimeout(5.0)
            data = sock_conn.recv(1024 * 8)
            sock_conn.close()

        port_string, pyzo_startscript_repr = data.decode('utf-8').strip().split(None, 1)
        fp_startscript = ast.literal_eval(pyzo_startscript_repr)
        return port_string, fp_startscript

    # os.environ['PYZO_PROCESS_EVENTS_WHILE_DEBUGGING'] = '1'

    shell_config = {
        'name': 'ext. shell',
        'gui': 'none',
        'pythonPath': '',
        'projectPath': '',
        'scriptFile': '',
        'startDir': '',
        'startupScript': textwrap.dedent("""
            # AFTER_GUI
            pass
        """.strip('\n')),
        'argv': '',
    }

    port_string, fp_startscript = get_pyzo_shell_interface(shell_config)

    cwd = os.path.dirname(os.path.dirname(fp_startscript))

    with open(fp_startscript, 'rt', encoding='utf-8') as fd:
        tt = fd.read()

    sys.path.append(cwd)
    sys.argv = [fp_startscript, port_string]
    exec(compile(tt, fp_startscript, 'exec'))


try:
    _run_pyzo_kernel()
except SystemExit:
    pass
finally:
    pass  # do some clean-up here if necessary



"""Module yoton.misc

Defines a few basic constants, classes and functions.

Most importantly, it defines a couple of specific buffer classes that
are used for the low-level messaging.

"""

import sys
import time
import struct
import socket
import threading
import random
from collections import deque


# Version dependent defs
V2 = sys.version_info[0] == 2
if V2:
    if sys.platform.startswith("java"):
        import __builtin__ as D  # Jython
    else:
        D = __builtins__
    if not isinstance(D, dict):
        D = D.__dict__
    bytes = D["str"]
    str = D["unicode"]
    xrange = D["xrange"]
    basestring = basestring  # noqa
    long = long  # noqa
else:
    basestring = str  # to check if instance is string
    bytes, str = bytes, str
    long = int  # for the port
    xrange = range


def getErrorMsg():
    """getErrorMsg()
    Return a string containing the error message. This is useful, because
    there is no uniform way to catch exception objects in Python 2.x and
    Python 3.x.
    """

    # Get traceback info
    type, value, tb = sys.exc_info()

    # Store for debugging?
    if True:
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb

    # Print
    err = ""
    try:
        if not isinstance(value, (OverflowError, SyntaxError, ValueError)):
            while tb:
                err = "line %i of %s." % (
                    tb.tb_frame.f_lineno,
                    tb.tb_frame.f_code.co_filename,
                )
                tb = tb.tb_next
    finally:
        del tb
    return str(value) + "\n" + err


def slot_hash(name):
    """slot_hash(name)

    Given a string (the slot name) returns a number between 8 and 2**64-1
    (just small enough to fit in a 64 bit unsigned integer). The number
    is used as a slot id.

    Slots 0-7 are reserved slots.

    """
    fac = 0xD2D84A61
    val = 0
    offset = 8
    for c in name:
        val += (val >> 3) + (ord(c) * fac)
    val += (val >> 3) + (len(name) * fac)
    return offset + (val % (2**64 - offset))


def port_hash(name):
    """port_hash(name)

    Given a string, returns a port number between 49152 and 65535.
    (2**14 (16384) different posibilities)
    This range is the range for dynamic and/or private ports
    (ephemeral ports) specified by iana.org.
    The algorithm is deterministic, thus providing a way to map names
    to port numbers.

    """
    fac = 0xD2D84A61
    val = 0
    for c in name:
        val += (val >> 3) + (ord(c) * fac)
    val += (val >> 3) + (len(name) * fac)
    return 49152 + (val % 2**14)


def split_address(address):
    """split_address(address) -> (protocol, hostname, port)

    Split address in protocol, hostname and port. The address has the
    following format: "protocol://hostname:port". If the protocol is
    omitted, TCP is assumed.

    The hostname is the name or ip-address of the computer to connect to.
    One can use "localhost" for a connection that bypasses some
    network layers (and is not visible from the outside). One can use
    "publichost" for a connection at the current computers IP address
    that is visible from the outside.

    The port can be an integer, or a string. In the latter case the integer
    port number is calculated using a hash. One can also use "portname+offset"
    to specify an integer offset for the port number.

    """

    # Check
    if not isinstance(address, basestring):
        raise ValueError("Address should be a string.")

    # Is the protocol explicitly defined (zeromq compatibility)
    protocol, _, address = address.rpartition("://")
    protocol = protocol.lower() or "tcp"

    # Split
    if ":" not in address:
        raise ValueError("Address should be in format 'host:port'.")
    host, port = address.split(":", 1)

    # Process host
    if host.lower() == "localhost":
        host = "127.0.0.1"

    ph = "publichost"
    if host.lower() == ph:
        host = ph + "0"

    if host.lower().startswith(ph) and set(host[len(ph) :]).issubset("0123456789"):
        index = int(host[len(ph) :])
        hostname = socket.gethostname()
        ipaddrlist = socket.gethostbyname_ex(hostname)[2]
        try:
            host = ipaddrlist[index]  # This resolves to 127.0.1.1 on some Linuxes
        except IndexError:
            raise ValueError("Invalid index (%i) in public host addresses." % index)

    # Process port
    try:
        port = int(port)
    except ValueError:
        # Convert to int, using a hash

        # Is there an offset?
        offset = 0
        if "+" in port:
            port, offset = port.split("+", 1)
            try:
                offset = int(offset)
            except ValueError:
                raise ValueError("Invalid offset in address")

        # Convert
        port = port_hash(port) + offset

    # Check port
    # if port < 1024 or port >= 2**16:
    #    raise ValueError("The port must be in the range [1024, 2^16>.")
    if port >= 2**16:
        raise ValueError("The port must be in the range [0, 2^16>.")

    # Done
    return protocol, host, port


class UID:
    """UID

    Represents an 8-byte (64 bit) Unique Identifier.

    """

    _last_timestamp = 0

    def __init__(self, id=None):
        # Create nr consisting of two parts
        if id is None:
            self._nr = self._get_time_int() << 32
            self._nr += self._get_random_int()
        elif isinstance(id, (int, long)):
            self._nr = id
        else:
            raise ValueError("The id given to UID() should be an int.")

    def __repr__(self):
        h = self.get_hex()
        return "<UID %s-%s>" % (h[:8], h[8:])

    def get_hex(self):
        """get_hex()

        Get the hexadecimal representation of this UID. The returned
        string is 16 characters long.

        """
        return format(self._nr, "016x")

    def get_bytes(self):
        """get_bytes()

        Get the UID as bytes.

        """
        return struct.pack("<Q", self._nr)

    def get_int(self):
        """get_int()

        Get the UID as a 64 bit (long) integer.

        """
        return self._nr

    def _get_random_int(self):
        return random.randrange(1 << 32)

    def _get_time_int(self):
        # Get time stamp in steps of milliseconds
        timestamp = int(time.time() * 1000)
        # Increase by one if the same as last time
        if timestamp <= UID._last_timestamp:
            timestamp = UID._last_timestamp + 1
        # Store for next time
        UID._last_timestamp = timestamp
        # Truncate to 4 bytes. If the time goes beyond the integer limit, we just
        # restart counting. With this setup, the cycle is almost 50 days
        timestamp &= 0xFFFFFFFF
        # Don't allow 0
        if timestamp == 0:
            timestamp += 1
            UID._last_timestamp += 1
        return timestamp


class PackageQueue(object):
    """PackageQueue(N, discard_mode='old')

    A queue implementation that can be used in blocking and non-blocking
    mode and allows peeking. The queue has a limited size. The user
    can specify whether old or new messages should be discarded.

    Uses a deque object for the queue and a threading.Condition for
    the blocking.

    """

    class Empty(Exception):
        def __init__(self):
            Exception.__init__(self, "pop from an empty PackageQueue")

        pass

    def __init__(self, N, discard_mode="old"):
        # Instantiate queue and condition
        self._q = deque()
        self._condition = threading.Condition()

        # Store max number of elements in queue
        self._maxlen = int(N)

        # Store discard mode as integer
        discard_mode = discard_mode.lower()
        if discard_mode == "old":
            self._discard_mode = 1
        elif discard_mode == "new":
            self._discard_mode = 2
        else:
            raise ValueError("Invalid discard mode.")

    def full(self):
        """full()

        Returns True if the number of elements is at its maximum right now.
        Note that in theory, another thread might pop an element right
        after this function returns.

        """
        return len(self) >= self._maxlen

    def empty(self):
        """empty()

        Returns True if the number of elements is zero right now. Note
        that in theory, another thread might add an element right
        after this function returns.

        """
        return len(self) == 0

    def push(self, x):
        """push(item)

        Add an item to the queue. If the queue is full, the oldest
        item in the queue, or the given item is discarded.

        """

        condition = self._condition
        with condition:
            q = self._q

            if len(q) < self._maxlen:
                # Add now and notify any waiting threads in get()
                q.append(x)
                condition.notify()  # code at wait() procedes
            else:
                # Full, either discard or pop (no need to notify)
                if self._discard_mode == 1:
                    q.popleft()  # pop old
                    q.append(x)
                elif self._discard_mode == 2:
                    pass  # Simply do not add

    def insert(self, x):
        """insert(x)

        Insert an item at the front of the queue. A call to pop() will
        get this item first. This should be used in rare circumstances
        to give an item priority. This method never causes items to
        be discarded.

        """

        condition = self._condition
        with condition:
            self._q.appendleft(x)
            condition.notify()  # code at wait() procedes

    def pop(self, block=True):
        """pop(block=True)

        Pop the oldest item from the queue. If there are no items in the
        queue:
          * the calling thread is blocked until an item is available
            (if block=True, default);
          * an PackageQueue.Empty exception is raised (if block=False);
          * the calling thread is blocked for 'block' seconds (if block
            is a float).

        """

        condition = self._condition
        with condition:
            q = self._q

            if not block:
                # Raise empty if no items in the queue
                if not len(q):
                    raise self.Empty()
            elif block is True:
                # Wait for notify (awakened does not guarantee len(q)>0)
                while not len(q):
                    condition.wait()
            elif isinstance(block, float):
                # Wait if no items, then raise error if still no items
                if not len(q):
                    condition.wait(block)
                    if not len(q):
                        raise self.Empty()
            else:
                raise ValueError("Invalid value for block in PackageQueue.pop().")

            # Return item
            return q.popleft()

    def peek(self, index=0):
        """peek(index=0)

        Get an item from the queue without popping it. index=0 gets the
        oldest item, index=-1 gets the newest item. Note that index access
        slows to O(n) time in the middle of the queue (due to the underlying
        deque object).

        Raises an IndexError if the index is out of range.

        """
        return self._q[index]

    def __len__(self):
        return self._q.__len__()

    def clear(self):
        """clear()

        Remove all items from the queue.

        """
        with self._condition:
            self._q.clear()


class TinyPackageQueue(PackageQueue):
    """TinyPackageQueue(N1, N2, discard_mode='old', timeout=1.0)

    A queue implementation that can be used in blocking and non-blocking
    mode and allows peeking. The queue has a tiny-size (N1). When this size
    is reached, a call to push() blocks for up to timeout seconds. The
    real size (N2) is the same as in the PackageQueue class.

    The tinysize mechanism can be used to semi-synchronize a consumer
    and a producer, while still having a small queue and without having
    the consumer fully block.

    Uses a deque object for the queue and a threading.Condition for
    the blocking.

    """

    def __init__(self, N1, N2, discard_mode="old", timeout=1.0):
        PackageQueue.__init__(self, N2, discard_mode)

        # Store limit above which the push() method will block
        self._tinylen = int(N1)

        # Store timeout
        self._timeout = timeout

    def push(self, x):
        """push(item)

        Add an item to the queue. If the queue has >= n1 values,
        this function will block timeout seconds, or until an item is
        popped from another thread.

        """

        condition = self._condition
        with condition:
            q = self._q
            lq = len(q)

            if lq < self._tinylen:
                # We are on safe side. Wake up any waiting threads if queue was empty
                q.append(x)
                condition.notify()  # pop() at wait() procedes
            elif lq < self._maxlen:
                # The queue is above its limit, but not full
                condition.wait(self._timeout)
                q.append(x)
            else:
                # Full, either discard or pop (no need to notify)
                if self._discard_mode == 1:
                    q.popleft()  # pop old
                    q.append(x)
                elif self._discard_mode == 2:
                    pass  # Simply do not add

    def pop(self, block=True):
        """pop(block=True)

        Pop the oldest item from the queue. If there are no items in the
        queue:
          * the calling thread is blocked until an item is available
            (if block=True, default);
          * a PackageQueue.Empty exception is raised (if block=False);
          * the calling thread is blocked for 'block' seconds (if block
            is a float).

        """

        condition = self._condition
        with condition:
            q = self._q

            if not block:
                # Raise empty if no items in the queue
                if not len(q):
                    raise self.Empty()
            elif block is True:
                # Wait for notify (awakened does not guarantee len(q)>0)
                while not len(q):
                    condition.wait()
            elif isinstance(block, float):
                # Wait if no items, then raise error if still no items
                if not len(q):
                    condition.wait(block)
                    if not len(q):
                        raise self.Empty()
            else:
                raise ValueError("Invalid value for block in TinyPackageQueue.pop().")

            # Notify if this pop would reduce the length below the threshold
            if len(q) <= self._tinylen:
                condition.notify_all()

            # Return item
            return q.popleft()

    def clear(self):
        """clear()

        Remove all items from the queue.

        """

        with self._condition:
            lq = len(self._q)
            self._q.clear()
            if lq >= self._tinylen:
                self._condition.notify()



"""Module logging

Functionality for logging in pyzo.

"""

import os
import sys
import time
import pyzo

# todo: enable logging to a file?

# Define prompts
try:
    sys.ps1  # noqa: B018
except AttributeError:
    sys.ps1 = ">>> "
try:
    sys.ps2  # noqa: B018
except AttributeError:
    sys.ps2 = "... "


class DummyStd:
    """For when std is not available."""

    def __init__(self):
        self._closed = False

    def write(self, text):
        pass

    def encoding(self):
        return "utf-8"

    @property
    def closed(self):
        return self._closed

    def close(self):
        self._closed = False

    def flush(self):
        pass


original_print = print


def print(*args, **kwargs):
    ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    preamble = ts + ":"
    if kwargs.get("sep", " ") != " ":
        preamble += " "
    original_print(preamble, *args, **kwargs)


def splitConsole(stdoutFun=None, stderrFun=None):
    """Splits the stdout and stderr streams.

    On each call to their write methods, in addition to the original
    write method being called, will call the given functions.
    Returns the history of the console (combined stdout and stderr).
    Used by the logger shell.
    """

    # Split stdout and stderr
    sys.stdout = OutputStreamSplitter(sys.stdout)
    sys.stderr = OutputStreamSplitter(sys.stderr)

    # Make them share their history
    sys.stderr._history = sys.stdout._history

    # Set defer functions
    if stdoutFun:
        sys.stdout._deferFunction = stdoutFun
    if stderrFun:
        sys.stderr._deferFunction = stderrFun

    # Return history
    return "".join(sys.stdout._history)


class OutputStreamSplitter:
    """This class is used to replace stdout and stderr output
    streams. It defers the stream to the original and to
    a function that can be registered.
    Used by the logger shell.
    """

    def __init__(self, fileObject):
        # Init, copy properties if it was already a splitter
        if isinstance(fileObject, OutputStreamSplitter):
            self._original = fileObject._original
            self._history = fileObject._history
            self._deferFunction = fileObject._deferFunction
        else:
            self._original = fileObject
            self._history = []
            self._deferFunction = self.dummyDeferFunction

        # Replace original with a dummy if None
        if self._original is None:
            self._original = DummyStd()

    def dummyDeferFunction(self, text):
        pass

    def write(self, text):
        """Write method."""
        self._original.write(text)
        self._history.append(text)
        if os.getenv("PYZO_LOG", ""):
            with open(os.getenv("PYZO_LOG"), "at") as f:
                f.write(text)
        try:
            self._deferFunction(text)
        except Exception:
            pass  # self._original.write("error writing to deferred stream")
        # Show in statusbar
        if pyzo.config.view.showStatusbar and len(text) > 1:
            if pyzo.main:
                pyzo.main.statusBar().showMessage(text, 5000)

    def flush(self):
        return self._original.flush()

    @property
    def closed(self):
        return self._original.closed

    def close(self):
        return self._original.close()

    def encoding(self):
        return self._original.encoding()


# Split now, with no defering
splitConsole()



"""Module manager

This module contains a static class that can be used for some
management tasks.

"""

import os
import sys

from .qt import QtGui, QtCore, QtWidgets  # noqa

Qt = QtCore.Qt

from . import parsers


class Manager:
    """Manager

    Static class to do some management tasks:
      * It manages the parsers
      * Getting style element descriptions of all parsers
      * Linking file extensions to parsers
      * Font information

    """

    _defaultFontFamily = "dummy_font_family_name"

    # Static dict of all parsers
    _parserInstances = {}
    _fileExtensions = {}
    _shebangKeywords = {}

    ## Parsers

    #     @classmethod
    #     def collectParsersDynamically(cls):
    #         """ insert the function is this module's namespace.
    #         """
    #
    #         # Get the path of this subpackage
    #         path = __file__
    #         path = os.path.dirname( os.path.abspath(path) )
    #
    #         # Determine if we're in a zipfile
    #         i = path.find('.zip')
    #         if i>0:
    #             # get list of files from zipfile
    #             path = path[:i+4]
    #             z = zipfile.ZipFile(path)
    #             files = [os.path.split(i)[-1] for i in z.namelist()
    #                         if 'codeeditor' in i and 'parsers' in i]
    #         else:
    #             # get list of files from file system
    #             files = os.listdir(path)
    #
    #         # Extract all parsers
    #         parserModules = []
    #         for file in files:
    #
    #             # Only python files
    #             if file.endswith('.pyc'):
    #                 if file[:-1] in files:
    #                     continue # Only try import once
    #             elif not file.endswith('.py'):
    #                 continue
    #             # Only syntax files
    #             if '_parser.' not in file:
    #                 continue
    #
    #             # Import module
    #             fullfile = os.path.join(path, file)
    #             modname = os.path.splitext(file)[0]
    #             print('modname', modname)
    #             mod = __import__("codeeditor.parsers."+modname, fromlist=[modname])
    #             parserModules.append(mod)
    #
    #         print(parserModules)

    @classmethod
    def _collectParsers(cls):
        """Collect all parser classes. This function is called on startup."""

        # Prepare (use a set to prevent duplicates)
        foundParsers = set()
        G = parsers.__dict__
        ModuleClass = os.__class__

        # Collect parser classes
        for module_name in G:
            # Check if it is indeed a module, and if it has the right name
            if not isinstance(G[module_name], ModuleClass):
                continue
            if not module_name.endswith("_parser"):
                continue
            # Collect all valid classes from the module
            moduleDict = G[module_name].__dict__
            for name_in_module in moduleDict:
                ob = moduleDict[name_in_module]
                if isinstance(ob, type) and issubclass(ob, parsers.Parser):
                    foundParsers.add(ob)

        # Put in list with the parser names as keys
        parserInstances = {}
        for parserClass in foundParsers:
            name = parserClass.getParserName()
            if name:
                # Try instantiating the parser
                try:
                    parserInstances[name] = parserInstance = parserClass()
                except Exception:
                    # We cannot get the exception object in a Python2/Python3
                    # compatible way
                    print('Could not instantiate parser "{}".'.format(name))
                    continue

                # Register extensions and shebang keywords for this parser
                for ext in parserInstance.filenameExtensions():
                    cls._fileExtensions[ext] = name
                for skw in parserInstance.shebangKeywords():
                    cls._shebangKeywords[skw] = name

        # Store
        cls._parserInstances = parserInstances

    @classmethod
    def getParserNames(cls):
        """Get a list of all available parsers."""
        return list(cls._parserInstances.keys())

    @classmethod
    def getParserByName(cls, parserName):
        """Get the parser object corresponding to the given name.

        If no parser is known by the given name, a warning message
        is printed and None is returned.
        """
        if not parserName:
            return parsers.Parser()  # Default dummy parser

        # Case insensitive
        parserName = parserName.lower()

        # Return instantiated parser object.
        if parserName in cls._parserInstances:
            return cls._parserInstances[parserName]
        else:
            print('Warning: no parser known by the name "{}".'.format(parserName))
            print("I know these: ", cls._parserInstances.keys())
            return parsers.Parser()  # Default dummy parser

    @classmethod
    def getStyleElementDescriptionsForAllParsers(cls):
        """Get all style element descriptions corresponding to
        the tokens of all parsers.

        This function is used by the code editor to register all syntax
        element styles to the code editor class.
        """
        descriptions = {}
        for parser in cls._parserInstances.values():
            for token in parser.getUsedTokens():
                description = token.description
                descriptions[description.key] = description

        return list(descriptions.values())

    ## File extensions

    @classmethod
    def suggestParserfromFilenameExtension(cls, ext):
        """Given a filename extension, returns the name of the suggested
        parser corresponding to the language of the file.

        See also registerFilenameExtension()
        """

        # Normalize ext
        ext = "." + ext.lstrip(".").lower()

        # Get parser
        if ext in cls._fileExtensions:
            return cls._fileExtensions[ext]
        else:
            return ""

    @classmethod
    def suggestParserfromText(cls, text):
        """Given a text, returns the name of the suggested
        parser corresponding to the language of the file.

        See also registerShebangKeyword()
        """
        shebangline = None
        for line in text[:1000].splitlines():
            line = line.strip()
            if line.startswith("#!"):
                shebangline = line
                break
        if shebangline is None:
            return ""
        shebangline = shebangline[2:].split()  # takes care of eventual space after #!
        if len(shebangline) == 0:
            return ""
        interpreter = os.path.basename(shebangline[0])
        if interpreter == "env" and len(shebangline) > 1:
            interpreter = shebangline[1]

        # Get parser
        if interpreter in cls._shebangKeywords:
            return cls._shebangKeywords[interpreter]
        else:
            return ""

    @classmethod
    def suggestParser(cls, ext, text):
        """Given a filename extension and text, returns the name of the suggested
        parser corresponding to the language of the file.

        See also registerFilenameExtension() and registerShebangKeyword()
        """
        parser = cls.suggestParserfromText(text)
        if parser == "":
            parser = cls.suggestParserfromFilenameExtension(ext)
        parser = cls.getParserByName(parser).disambiguate(text)
        return parser

    @classmethod
    def registerFilenameExtension(cls, ext, parser):
        """Registers the given filename extension to the given parser.
        The parser can be a Parser instance or its name.

        This function can be used to register extensions to parsers
        that are not registered by default.

        """
        # Normalize ext
        ext = "." + ext.lstrip(".").lower()
        # Check parser
        if isinstance(parser, parsers.Parser):
            parser = parser.name()
        # Register
        cls._fileExtensions[ext] = parser

    @classmethod
    def registerShebangKeyword(cls, shebangKeyword, parser):
        """Registers the given shebang keyword (interpreter) to the given parser.
        The parser can be a Parser instance or its name.

        This function can be used to register shebang keywords to parsers
        that are not registered by default.

        """
        # Check parser
        if isinstance(parser, parsers.Parser):
            parser = parser.name()
        # Register
        cls._shebangKeywords[shebangKeyword] = parser

    ## Fonts

    @classmethod
    def fontNames(cls):
        """Get a list of all monospace fonts available on this system."""
        db = QtGui.QFontDatabase
        return [fn for fn in db.families() if db.isFixedPitch(fn)]

    @classmethod
    def setDefaultFontFamily(cls, name):
        """Set the default (monospace) font family name for this system.

        This should be set only once during startup.
        """
        cls._defaultFontFamily = name

    @classmethod
    def defaultFont(cls):
        """Get the default (monospace) font for this system. Returns a QFont object."""

        # Get font family
        f = QtGui.QFont(cls._defaultFontFamily)
        f.setStyleHint(f.StyleHint.TypeWriter, f.StyleStrategy.PreferDefault)
        fi = QtGui.QFontInfo(f)
        family = fi.family()

        # Get the font size
        size = 9
        if sys.platform.startswith("darwin"):
            # Account for Qt font size difference
            # http://qt-project.org/forums/viewthread/27201
            # Win/linux use 96 ppi, OS X uses 72 -> 133% ratio
            size = int(size * 1.33333 + 0.4999)

        # Done
        return QtGui.QFont(family, size)


# Init
try:
    Manager._collectParsers()
except Exception as why:
    print("Error collecting parsers")
    print(why)
